# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines utility functions."""

from typing import Iterable, Union, Optional, Any

from jax.tree_util import register_pytree_node
from jax.lib import xla_bridge
import jax.numpy as jnp
from jax import jit

from functools import partial

import numpy as onp

Array = jnp.ndarray

i16 = jnp.int16
i32 = jnp.int32
i64 = jnp.int64

f32 = jnp.float32
f64 = jnp.float64


def static_cast(*xs):
  """Function to cast a value to the lowest dtype that can express it."""
  # NOTE(schsam): static_cast is so named because it cannot be jit.
  if xla_bridge.get_backend().platform == 'tpu':
    return (jnp.array(x, jnp.float32) for x in xs)
  else:
    return (jnp.array(x, dtype=onp.min_scalar_type(x)) for x in xs)


def register_pytree_namedtuple(cls):
  register_pytree_node(
      cls,
      lambda xs: (tuple(xs), None),
      lambda _, xs: cls(*xs))


def merge_dicts(a, b, ignore_unused_parameters=False):
  if not ignore_unused_parameters:
    return {**a, **b}

  merged = dict(a)
  for key in merged.keys():
    b_val = b.get(key)
    if b_val is not None:
      merged[key] = b_val
  return merged


@partial(jit, static_argnums=(1,))
def safe_mask(mask, fn, operand, placeholder=0):
  masked = jnp.where(mask, operand, 0)
  return jnp.where(mask, fn(masked), placeholder)


def high_precision_sum(X: Array,
                       axis: Optional[Union[Iterable[int], int]]=None,
                       keepdims: bool=False):
  """Sums over axes at 64-bit precision then casts back to original dtype."""
  return jnp.array(
      jnp.sum(X, axis=axis, dtype=f64, keepdims=keepdims), dtype=X.dtype)


def maybe_downcast(x):
  if isinstance(x, jnp.ndarray) and x.dtype is jnp.dtype('float64'):
    return x
  return jnp.array(x, f32)


def is_array(x: Any) -> bool:
  return isinstance(x, (jnp.ndarray, onp.ndarray))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to calculate the elastic modulus tensor for athermal systems.

  The elastic modulus tensor describes a material's response to different
  boundary deformations. Specifically, for a small deformation given by a
  symmetric strain tensor e, the change in energy is

  .. math::

     U / V^0 = U^0/V^0 + s^0_{ij} e_{ji} + (1/2) C_{ijkl} e_{ij} e_{kl} + \ldots{}
  
  where :math:`V^0` is the volume, :math:`U^0` is the initial energy, :math:`s^0` is the residual stress
  tensor of the undeformed system, and :math:`C` is the elastic modulus tensor. :math:`C` is
  a fourth-rank tensor of shape (dimension,dimension,dimension,dimension), with
  the following symmetries.

    Minor symmetries:

    .. math::

       C_{ijkl} = C_{jikl} = C_{ijlk}

    Major symmetries:

    .. math::

       C_{ijkl} = C_{lkij}

  The minor symmetries are also reflected in the symmetric nature
  of stress and strain tensors:

  .. math::
  
     s_{ij} = s_{ji}
     e_{ij} = e_{ji}

  In general, there are 21 independent elastic constants in 3 dimension (6 in 2
  dimensions). While systems with additional symmetries (e.g. isotropic,
  orthotropic, etc.) can be expressed with fewer constants, we do not assume any
  such additional symmetries.

  At zero temperature, the response of every particle to a deformation can be
  calculated explicitly to linear order, enabling the exact calculation of
  the elastic modulus tensor without the need of any finite differences or any
  approximations.

"""




from functools import partial
from typing import Dict, Callable, Union
from absl import logging

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jax import jit, lax, grad, vmap, jacfwd, jacrev, jvp

from jax_md import quantity
from jax_md.util import Array
from jax_md.util import f32
from jax_md.util import f64

def _get_strain_tensor_list(dim, dtype) -> Array:
  if dim == 2:
    strain_tensors = jnp.array([[[1, 0],[0, 0]],
                                [[0, 0],[0, 1]],
                                [[0, 1],[1, 0]],
                                [[1, 0],[0, 1]],
                                [[1, 1],[1, 0]],
                                [[0, 1],[1, 1]]], dtype=dtype)
  elif dim == 3:
    strain_tensors = jnp.array([[[1, 0, 0], [0, 0, 0], [0, 0, 0]],
                                [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
                                [[0, 0, 0], [0, 0, 0], [0, 0, 1]],
                                [[0, 0, 0], [0, 0, 1], [0, 1, 0]],
                                [[0, 0, 1], [0, 0, 0], [1, 0, 0]],
                                [[0, 1, 0], [1, 0, 0], [0, 0, 0]],
                                [[0, 0, 0], [0, 1, 0], [0, 0, 1]],
                                [[1, 0, 0], [0, 0, 0], [0, 0, 1]],
                                [[1, 0, 0], [0, 1, 0], [0, 0, 0]],
                                [[1, 0, 0], [0, 0, 1], [0, 1, 0]],
                                [[1, 0, 1], [0, 0, 0], [1, 0, 0]],
                                [[1, 1, 0], [1, 0, 0], [0, 0, 0]],
                                [[0, 0, 0], [0, 1, 1], [0, 1, 0]],
                                [[0, 0, 1], [0, 1, 0], [1, 0, 0]],
                                [[0, 1, 0], [1, 1, 0], [0, 0, 0]],
                                [[0, 0, 0], [0, 0, 1], [0, 1, 1]],
                                [[0, 0, 1], [0, 0, 0], [1, 0, 1]],
                                [[0, 1, 0], [1, 0, 0], [0, 0, 1]],
                                [[0, 0, 1], [0, 0, 1], [1, 1, 0]],
                                [[0, 1, 0], [1, 0, 1], [0, 1, 0]],
                                [[0, 1, 1], [1, 0, 0], [1, 0, 0]]], dtype=dtype)
  else:
    raise AssertionError('not implemented for {} dimensions'.format(dim))
  return strain_tensors

def _convert_responses_to_elastic_constants(response_all: Array) -> Array:
  if response_all.shape[0] == 6:
    cxxxx = response_all[0]
    cyyyy = response_all[1]
    cxyxy = 0.25 * response_all[2]
    cxxyy = 0.5  * (response_all[3] - cxxxx - cyyyy)
    cxxxy = 0.25 * (response_all[4] - 4 * cxyxy - cxxxx)
    cyyxy = 0.25 * (response_all[5] - 4 * cxyxy - cyyyy)

    C = jnp.array(
        [[[[cxxxx, cxxxy], [cxxxy, cxxyy]],
          [[cxxxy, cxyxy], [cxyxy, cyyxy]]],
         [[[cxxxy, cxyxy], [cxyxy, cyyxy]],
          [[cxxyy, cyyxy], [cyyxy, cyyyy]]]])

  elif response_all.shape[0] == 21:
    cxxxx = response_all[0];
    cyyyy = response_all[1];
    czzzz = response_all[2];
    cyzyz = response_all[3] / 4.;
    cxzxz = response_all[4] / 4.;
    cxyxy = response_all[5] / 4.;
    cyyzz = (response_all[6] - cyyyy - czzzz) / 2.;
    cxxzz = (response_all[7] - cxxxx - czzzz) / 2.;
    cxxyy = (response_all[8] - cxxxx - cyyyy) / 2.;
    cxxyz = (response_all[9] - cxxxx - 4. * cyzyz) / 4.;
    cxxxz = (response_all[10] - cxxxx - 4. * cxzxz) / 4.;
    cxxxy = (response_all[11] - cxxxx - 4. * cxyxy) / 4.;
    cyyyz = (response_all[12] - cyyyy - 4. * cyzyz) / 4.;
    cyyxz = (response_all[13] - cyyyy - 4. * cxzxz) / 4.;
    cyyxy = (response_all[14] - cyyyy - 4. * cxyxy) / 4.;
    czzyz = (response_all[15] - czzzz - 4. * cyzyz) / 4.;
    czzxz = (response_all[16] - czzzz - 4. * cxzxz) / 4.;
    czzxy = (response_all[17] - czzzz - 4. * cxyxy) / 4.;
    cyzxz = (response_all[18] - 4. * cyzyz - 4. * cxzxz) / 8.;
    cyzxy = (response_all[19] - 4. * cyzyz - 4. * cxyxy) / 8.;
    cxzxy = (response_all[20] - 4. * cxzxz - 4. * cxyxy) / 8.;

    C = jnp.array(
         [[[[cxxxx, cxxxy, cxxxz],
            [cxxxy, cxxyy, cxxyz],
            [cxxxz, cxxyz, cxxzz]],
           [[cxxxy, cxyxy, cxzxy],
            [cxyxy, cyyxy, cyzxy],
            [cxzxy, cyzxy, czzxy]],
           [[cxxxz, cxzxy, cxzxz],
            [cxzxy, cyyxz, cyzxz],
            [cxzxz, cyzxz, czzxz]]],
          [[[cxxxy, cxyxy, cxzxy],
            [cxyxy, cyyxy, cyzxy],
            [cxzxy, cyzxy, czzxy]],
           [[cxxyy, cyyxy, cyyxz],
            [cyyxy, cyyyy, cyyyz],
            [cyyxz, cyyyz, cyyzz]],
           [[cxxyz, cyzxy, cyzxz],
            [cyzxy, cyyyz, cyzyz],
            [cyzxz, cyzyz, czzyz]]],
          [[[cxxxz, cxzxy, cxzxz],
            [cxzxy, cyyxz, cyzxz],
            [cxzxz, cyzxz, czzxz]],
           [[cxxyz, cyzxy, cyzxz],
            [cyzxy, cyyyz, cyzyz],
            [cyzxz, cyzyz, czzyz]],
           [[cxxzz, czzxy, czzxz],
            [czzxy, cyyzz, czzyz],
            [czzxz, czzyz, czzzz]]]])
  else:
    raise AssertionError('response_all has incorrect shape')
  return C

def athermal_moduli(energy_fn: Callable[..., Array],
                    tether_strength: float=1e-10,
                    gradient_check: Array=None,
                    cg_tol: float=1e-7,
                    check_convergence: bool=False
                    ) -> Callable[..., Array]:
  """ Setup calculation of elastic modulus tensor.

  Args:
    energy_fn: A function that computes the energy of the system. This
      function must take as an argument `perturbation` which perturbs the
      box shape. Any energy function constructed using `smap` or in `energy.py`
      with a standard space will satisfy this property.
    tether_strength: Scalar. Strength of the "tether" applied to each particle,
      which can be necessary to make the Hessian matrix non-singular. Solving
      for the non-affine response of each particle requires that the Hessian is
      positive definite. However, there can often be zero modes (eigenvectors of
      the Hessian with zero eigenvalue) that do not couple to the boundary, and
      therefore do not affect the elastic constants despite the zero eigenvalue.
      The most common example is the global translational modes. To solve for
      the non-affine response, we consider the "tethered Hessian"
      `H + tether_strength * I`, 
      where `I` is the identity matrix and `tether_strength` is a small constant.
    gradient_check: `None` or scalar. If not `None`, a check will be performed
      to guarantee that the maximum component of the gradient is less than
      `gradient_check`. In other words, that
      `jnp.amax(jnp.abs(grad(energy_fn)(R, box=box))) < gradient_check == True`
      NOTE: JAX currently does not support proper runtime error handling.
      Therefore, if this check fails, the calculation will return an array
      of `jnp.nan`'s. It is the users responsibility, if they want to use this
      check, to then ensure that the returned array is not full of nans.
    cg_tol: scalar. Tolerance used when solving for the non-affine response.
    check_convergence: bool. If true, `calculate_EMT` will return a boolean
      flag specifying if the cg solve routine converged to the desired
      tolerance. The default is `False`, but convergence checking is highly
      recommended especially when using 32-bit precision data.

  Return: A function to calculate the elastic modulus tensor
  """

  def calculate_emt(R: Array,
                    box: Array,
                    **kwargs) -> Array:
    """Calculate the elastic modulus tensor.

    `energy_fn(R)` corresponds to the state around which we are expanding

    Args:
      R: array of shape `[N,dimension]` of particle positions. This does not
        generalize to arbitrary dimensions and is only implemented for
        `dimension == 2` and `dimension == 3`.
      box: A box specifying the shape of the simulation volume. Used to infer
        the volume of the unit cell.

    Return: C or the tuple (C,converged)
      where C is the Elastic modulus tensor as an array of shape `[dimension,
      dimension,dimension,dimension]` that respects the major and minor
      symmetries, and converged is a boolean flag (see above).

    """
    if not (R.shape[-1] == 2 or R.shape[-1] == 3):
      raise AssertionError('Only implemented for 2d and 3d systems.')

    if R.dtype is not jnp.dtype('float64'):
      logging.warning('Elastic modulus calculations can sometimes lose '
                      'precision when not using 64-bit precision.')

    dim = R.shape[-1]

    def setup_energy_fn_general(strain_tensor):
      I = jnp.eye(dim, dtype=R.dtype)
      @jit
      def energy_fn_general(R, gamma):
        perturbation = I + gamma * strain_tensor
        return energy_fn(R, perturbation=perturbation, **kwargs)
      return energy_fn_general

    def get_affine_response(strain_tensor):
      energy_fn_general = setup_energy_fn_general(strain_tensor)
      d2U_dRdgamma = jacfwd(jacrev(energy_fn_general,argnums=0),argnums=1)(R,0.)
      d2U_dgamma2  = jacfwd(jacrev(energy_fn_general,argnums=1),argnums=1)(R,0.)
      return d2U_dRdgamma, d2U_dgamma2

    strain_tensors = _get_strain_tensor_list(dim, R.dtype)
    d2U_dRdgamma_all,d2U_dgamma2_all = vmap(get_affine_response)(strain_tensors)

    #Solve the system of equations.
    energy_fn_Ronly = partial(energy_fn, **kwargs)
    def hvp(f, primals, tangents):
      return jvp(grad(f), primals, tangents)[1]
    def hvp_specific_with_tether(v):
      return hvp(energy_fn_Ronly, (R,), (v,)) + tether_strength * v

    non_affine_response_all = jsp.sparse.linalg.cg(
        vmap(hvp_specific_with_tether),
        d2U_dRdgamma_all,
        tol=cg_tol
        )[0]
    #The above line should be functionally equivalent to:
    #H0=hessian(energy_fn)(R, box=box, **kwargs).reshape(R.size,R.size) \
    #    + tether_strength * jnp.identity(R.size)
    #non_affine_response_all = jnp.transpose(jnp.linalg.solve(
    #   H0,
    #   jnp.transpose(d2U_dRdgamma_all))
    #   )

    residual = jnp.linalg.norm(vmap(hvp_specific_with_tether)(
        non_affine_response_all) - d2U_dRdgamma_all
      )
    converged = residual / jnp.linalg.norm(d2U_dRdgamma_all) < cg_tol

    response_all = d2U_dgamma2_all - jnp.einsum("nij,nij->n",
                                                d2U_dRdgamma_all,
                                                non_affine_response_all)

    vol_0 = quantity.volume(dim, box)
    response_all = response_all / vol_0
    C = _convert_responses_to_elastic_constants(response_all)

    # JAX does not allow proper runtime error handling in jitted function.
    # Instead, if the user requests a gradient check and the check fails,
    # we convert C into jnp.nan's. While this doesn't raise an exception,
    # it at least is very "loud".
    if gradient_check is not None:
      maxgrad = jnp.amax(jnp.abs(grad(energy_fn)(R, **kwargs)))
      C = lax.cond(maxgrad > gradient_check,
                   lambda _: jnp.nan * C,
                   lambda _: C,
                   None)

    if check_convergence:
      return C, converged
    else:
      return C

  return calculate_emt


def _get_mandel_mapping_weight(dim, dtype):
  if dim == 2:
    m_map  = jnp.array([[0,0],[1,1],[0,1]], dtype=jnp.int8)
    weight = jnp.array([1,1,jnp.sqrt(2)], dtype=dtype)
    return m_map, weight
  elif dim == 3:
    m_map  = jnp.array([[0,0],[1,1],[2,2],[1,2],[0,2],[0,1]], dtype=jnp.int8)
    weight = jnp.array([1,1,1,jnp.sqrt(2),jnp.sqrt(2),jnp.sqrt(2)], dtype=dtype)
    return m_map, weight
  else:
    raise AssertionError('dim must be 2 or 3')

def tensor_to_mandel(T: Array) -> Array:
  """ Convert a tensor to Mandel notation.

  Mandel notation is a way to represent symmetric second-rank tensors and
  fourth-rank tensors with minor symmetries in a reduced form. Pairs of indices
  are combined as follows:

      For tensors of shape (2,2) or (2,2,2,2):
      tensor indices        Mandel indices
      0,0 -------------->   0
      1,1 -------------->   1
      0,1 or 1,0 ------->   2

      For tensors of shape (3,3) or (3,3,3,3):
      tensor indices        Mandel indices
      0,0 -------------->   0
      1,1 -------------->   1
      2,2 -------------->   2
      1,2 or 2,1 ------->   3
      0,2 or 2,0 ------->   4
      0,1 or 1,0 ------->   5

  If mandel_index(i,j) performs the above index mapping, then the input T and
  output M satisfy
    M[mandel_index(i,j)] = T[i,j] * w
  or
    M[mandel_index(i,j), mandel_index(k,l)] = T[i,j,k,l] * w(i,j) * w(k,l)
  where
    w(i,j) = 1       if i==j
           = sqrt(2) if i!=j
  is a weight that is used to ensure proper contraction rules. Here (and only
  here) we do not assume major symmetries in fourth-rank tensors.

  Args:
    T: Array with 4 possible shapes:
      1. T.shape == (2,2)
         Convert a symmetric array of shape (2,2) to an array of shape (3,)
      2. T.shape == (3,3)
         Convert a symmetric array of shape (3,3) to an array of shape (6,)
      3. T.shape == (2,2,2,2)
         Convert a tensor of shape (2,2,2,2) with minor symmetries to an array
         of shape (3,3)
      4. T.shape == (3,3,3,3)
         Convert a tensor of shape (3,3,3,3) with minor symmetries to an array
         of shape (6,6)

  Output: Array of shape (3,), (6,), (3,3), or (6,6)

  see: https://sbrisard.github.io/janus/mandel.html (accessed 21 April, 2021)
  """
  dim = T.shape[0]
  if not (dim ==2 or dim == 3):
    raise AssertionError('dim must be 2 or 3')

  rank = len(T.shape)
  if not (rank == 2 or rank == 4):
    raise AssertionError('T must have rank 2 or 4')

  m_map, weight = _get_mandel_mapping_weight(dim, T.dtype)

  if rank == 2:
    extract = lambda idx, w: T[idx[0], idx[1]] * w
    M = vmap(extract, in_axes=(0,0))(m_map, weight)
  else:
    extract = lambda idx0, idx1, w0, w1: T[
        idx0[0], idx0[1], idx1[0], idx1[1]
        ] * w0 * w1
    M = vmap(
          vmap(extract,
            in_axes=(0,None,0,None)),
          in_axes=(None,0,None,0))(m_map, m_map, weight, weight)
  return M

def mandel_to_tensor(M: Array) -> Array:
  """ Perform the inverse of M = tensor_to_mandel(T).

  Args:
    M: Array of shape (3,), (6,), (3,3), or (6,6)

  Output: Array of shape (2,2), (3,3), (2,2,2,2), or (3,3,3,3)
  """
  mandel_dim = M.shape[0]
  if not (mandel_dim == 3 or mandel_dim == 6):
    raise AssertionError('M.shape[0] must be 3 or 6')

  rank = len(M.shape)
  if not (rank == 1 or rank == 2):
    raise AssertionError('T must have rank 1 or 2')

  def mandel_index(i, j):
    return lax.cond(i==j,
                    lambda ij: ij[0],
                    lambda ij: mandel_dim - ij[0] - ij[1],
                    (i,j))

  if mandel_dim == 3:
    dimension = 2
  else:
    dimension = 3

  tensor_range = jnp.arange(dimension)
  _, weight = _get_mandel_mapping_weight(dimension, M.dtype)

  if rank == 1:
    def extract(i, j):
      idx = mandel_index(i, j)
      return M[idx] / weight[idx]
    T = vmap(
          vmap(extract,
            in_axes=(0,None)),
          in_axes=(None,0))(tensor_range, tensor_range)
  else:
    def extract(i, j, k, l):
      idx0 = mandel_index(i, j)
      idx1 = mandel_index(k, l)
      return M[idx0, idx1] / (weight[idx0] * weight[idx1])
    T = vmap(
          vmap(
            vmap(
              vmap(extract,
                in_axes=(0,None,None,None)),
              in_axes=(None,0,None,None)),
            in_axes=(None,None,0,None)),
          in_axes=(None,None,None,0))(tensor_range,
                                      tensor_range,
                                      tensor_range,
                                      tensor_range)
  return T



@partial(jit,static_argnums=(1,))
def _extract_elements(C, as_dict):
  if C.shape[0] == 2:
    indices = ( [0, 1, 0, 0, 0, 0],
                [0, 1, 1, 0, 0, 1],
                [0, 1, 0, 1, 0, 1],
                [0, 1, 1, 1, 1, 1])
    clist = C[ indices ]
    if as_dict:
      names = ['cxxxx','cyyyy','cxyxy','cxxyy','cxxxy','cyyxy']
      return dict(zip(names, clist))
    else:
      return clist

  elif C.shape[0] == 3:
    indices = ( [0, 1, 2, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 1, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 1, 2, 1, 2, 2, 1, 2, 1, 1],
                [0, 1, 2, 1, 0, 0, 2, 2, 1, 1, 0, 0, 1, 1, 1, 2, 2, 2, 1, 1, 0],
                [0, 1, 2, 2, 2, 1, 2, 2, 1, 2, 2, 1, 2, 1, 1, 2, 2, 2, 2, 2, 2])
    clist = C[ indices ]
    if as_dict:
      names = ['cxxxx','cyyyy','czzzz','cyzyz','cxzxz','cxyxy','cyyzz','cxxzz',
               'cxxyy','cxxyz','cxxxz','cxxxy','cyyyz','cyyxz','cyyxy','czzyz',
               'czzxz','czzxy','cyzxz','cyzxy','cxzxy']
      return dict(zip(names, clist))
    else:
      return clist
  else:
    raise AssertionError('C has wrong shape')

def extract_elements(C: Array) -> Dict:
  """ Convert an elastic modulus tensor into a list of unique elements.

      In 2d, these are:
      cxxxx,cyyyy,cxyxy,cxxyy,cxxxy,cyyxy

      In 3d, these are:
      cxxxx,cyyyy,czzzz,cyzyz,cxzxz,cxyxy,cyyzz,cxxzz,cxxyy,cxxyz,cxxxz,cxxxy,
      cyyyz,cyyxz,cyyxy,czzyz,czzxz,czzxy,cyzxz,cyzxy,cxzxy

  Args:
    C: A previously calculated elastic modulus tensor represented as an
      array of shape (spatial_dimension,spatial_dimension,spatial_dimension,
      spatial_dimension), where spatial_dimension is either 2 or 3. C must
      satisfy both the major and minor symmetries, but this is not checked.
  Return: a dict of the 6 (21) unique elastic constants in 2 (3) dimensions.
  """
  return _extract_elements(C,True)

def extract_isotropic_moduli(C: Array) -> Dict:
  """ Extract commonly used isotropic constants.

  There are a number of important constants used to describe the linear
  elastic behavior of isotropic systems, including the bulk modulus, B,
  the shear modulus, G, the longitudinal modulus, M, the Young's modulus,
  E, and the Poisson's ratio, nu. This convenience function extracts them
  from an elastic modulus tensor C.

  Angle averaged quantities: While these quantities are defined for
  isotropic systems, one can still define an "angle-averaged shear
  modulus", for example, that averages over all possible shear
  deformations. This can be useful for systems that are statistically
  isotropic but where a particular realization is slightly anisotropic.
  The precise definitions are as follows:

  First, we define the "response", R, to a certain strain tensor e to be
    R = 2 * (U / V^0 - U^0/V^0 - s^0_{ij} e_{ji}) = C_{ijkl} e_{ij} e_{kl}

  Bulk modulus, B:
  This is the response to the rotationally invariant strain tensor:
      e = (1/2) * ( 1 0 )     or      e = (1/3) * ( 1 0 0 )
                  ( 0 1 )                         ( 0 1 0 )
                                                  ( 0 0 1 )

  Shear modulus, G:
  This is the response to the strain tensor:
      e = (1/2) * ( 0 1 )     or      e = (1/2) * ( 0 1 0 )
                  ( 1 0 )                         ( 1 0 0 )
                                                  ( 0 0 0 )
  averaged over all possible orientations. For perfectly isotropic
  systems, it should be equal to C[0,1,0,1].

  Longitudinal modulus, M:
  This is the response to the strain tensor:
      e = ( 1 0 )     or      e = ( 1 0 0 )
          ( 0 0 )                 ( 0 0 0 )
                                  ( 0 0 0 )
  averaged over all possible orientations. For perfectly isotropic
  systems, it should be equal to C[0,0,0,0].

  Young's modulus, E:
  This is a measure of tensile stiffness and is calculated a using well-
  known expression in terms of B and G.

  Poisson's ratio:
  This is a measure of deformations in directions perpendicular to an
  applied load and is calculated a using well-known expression in terms
  of B and G.

  Args:
    C: A previously calculated elastic modulus tensor represented as an
      array of shape (spatial_dimension,spatial_dimension,spatial_dimension,
      spatial_dimension), where spatial_dimension is either 2 or 3. C must
      satisfy both the major and minor symmetries, but this is not checked.

  Return: a dictionary containing the elastic constants.
  """
  if C.shape[0] == 2:
    cxxxx,cyyyy,cxyxy,cxxyy,cxxxy,cyyxy = _extract_elements(C,False)
    B = (cxxxx + cyyyy + 2. * cxxyy) / 4.
    G = (4. * cxyxy + cxxxx + cyyyy - 2. * cxxyy) / 8.
    M = B + G
    E = 4 * B * G / (B + G)
    nu = (B - G) / (B + G)

  elif C.shape[0] == 3:
    cxxxx,cyyyy,czzzz,cyzyz,cxzxz,cxyxy,cyyzz,cxxzz,cxxyy, \
    cxxyz,cxxxz,cxxxy,cyyyz,cyyxz,cyyxy,czzyz,czzxz,czzxy, \
    cyzxz,cyzxy,cxzxy = _extract_elements(C,False)
    B = (cxxxx + 2 * cxxyy + 2 * cxxzz + cyyyy + 2 * cyyzz + czzzz) / 9.
    G = (cxxxx - cxxyy - cxxzz + 3 * cxyxy + 3 * cxzxz + cyyyy - cyyzz
         + 3 * cyzyz + czzzz) / 15.
    M = B + 4 * G / 3
    E = 9 * B * G / (3 * B + G)
    nu = (3 * B - 2 * G) / (2 * (3 * B + G))
  else:
    raise AssertionError('C has incorrect shape')

  return {'B':B, 'G':G, 'M':M, 'E':E, 'nu':nu}










# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Spaces in which particles are simulated.

Spaces are pairs of functions containing:
  `displacement_fn(Ra, Rb, **kwargs)`:
    Computes displacements between pairs of particles. `Ra` and `Rb` should
    be ndarrays of shape `[spatial_dim]`. Returns an ndarray of shape `[spatial_dim]`. 
    To compute the displacement over more than one particle at a time see the 
    :meth:`map_product`, :meth:`map_bond`, and :meth:`map_neighbor` functions.
  `shift_fn(R, dR, **kwargs)`:
    Moves points at position `R` by an amount `dR`.

Spaces can accept keyword arguments allowing the space to be changed over the
course of a simulation. For an example of this use see :meth:`periodic_general`.

Although displacement functions are compute the displacement between two
points, it is often useful to compute displacements between multiple particles
in a vectorized fashion. To do this we provide three functions: `map_product`,
`map_bond`, and `map_neighbor`:
  map_product:
    Computes displacements between all pairs of points such that if
    `Ra` has shape `[n, spatial_dim]` and `Rb` has shape `[m, spatial_dim]` then the
    output has shape `[n, m, spatial_dim]`.
  map_bond:
    Computes displacements between all points in a list such that if
    `Ra` has shape `[n, spatial_dim]` and `Rb` has shape `[m, spatial_dim]` then the
    output has shape `[n, spatial_dim]`.
  map_neighbor:
    Computes displacements between points and all of their
    neighbors such that if `Ra` has shape `[n, spatial_dim]` and `Rb` has shape
    `[n, neighbors, spatial_dim]` then the output has shape
    `[n, neighbors, spatial_dim]`.
"""

from typing import Callable, Union, Tuple, Any, Optional

from jax.abstract_arrays import ShapedArray

from jax import eval_shape
from jax import vmap
from jax import custom_jvp

import jax

import jax.numpy as jnp

from jax_md.util import Array
from jax_md.util import f32
from jax_md.util import f64
from jax_md.util import safe_mask


# Types


DisplacementFn = Callable[[Array, Array], Array]
MetricFn = Callable[[Array, Array], float]
DisplacementOrMetricFn = Union[DisplacementFn, MetricFn]

ShiftFn = Callable[[Array, Array], Array]

Space = Tuple[DisplacementFn, ShiftFn]
Box = Array


# Exceptions


class UnexpectedBoxException(Exception):
  pass


# Primitive Spatial Transforms


def inverse(box: Box) -> Box:
  """Compute the inverse of an affine transformation."""
  if jnp.isscalar(box) or box.size == 1:
    return 1 / box
  elif box.ndim == 1:
    return 1 / box
  elif box.ndim == 2:
    return jnp.linalg.inv(box)
  raise ValueError(('Box must be either: a scalar, a vector, or a matrix. '
                    f'Found {box}.'))


def _get_free_indices(n: int) -> str:
  return ''.join([chr(ord('a') + i) for i in range(n)])


def raw_transform(box: Box, R: Array) -> Array:
  """Apply an affine transformation to positions.

  See `periodic_general` for a description of the semantics of `box`.

  Args:
    box: An affine transformation described in `periodic_general`.
    R: Array of positions. Should have  shape `(..., spatial_dimension)`.

  Returns:
    A transformed array positions of shape `(..., spatial_dimension)`.
  """
  if jnp.isscalar(box) or box.size == 1:
    return R * box
  elif box.ndim == 1:
    indices = _get_free_indices(R.ndim - 1) + 'i'
    return jnp.einsum(f'i,{indices}->{indices}', box, R)
  elif box.ndim == 2:
    free_indices = _get_free_indices(R.ndim - 1)
    left_indices = free_indices + 'j'
    right_indices = free_indices + 'i'
    return jnp.einsum(f'ij,{left_indices}->{right_indices}', box, R)
  raise ValueError(('Box must be either: a scalar, a vector, or a matrix. '
                    f'Found {box}.'))


@custom_jvp
def transform(box: Box, R: Array) -> Array:
  """Apply an affine transformation to positions.

  See `periodic_general` for a description of the semantics of `box`.

  Args:
    box: An affine transformation described in `periodic_general`.
    R: Array of positions. Should have  shape `(..., spatial_dimension)`.

  Returns:
    A transformed array positions of shape `(..., spatial_dimension)`.
  """
  return raw_transform(box, R)


@transform.defjvp
def transform_jvp(primals, tangents):
  box, R = primals
  dbox, dR = tangents
  return (transform(box, R), dR + transform(dbox, R))


def pairwise_displacement(Ra: Array, Rb: Array) -> Array:
  """Compute a matrix of pairwise displacements given two sets of positions.

  Args:
    Ra: Vector of positions; `ndarray(shape=[spatial_dim])`.
    Rb: Vector of positions; `ndarray(shape=[spatial_dim])`.

  Returns:
    Matrix of displacements; `ndarray(shape=[spatial_dim])`.
  """
  if len(Ra.shape) != 1:
    msg = (
      'Can only compute displacements between vectors. To compute '
      'displacements between sets of vectors use vmap or TODO.'
    )
    raise ValueError(msg)

  if Ra.shape != Rb.shape:
    msg = 'Can only compute displacement between vectors of equal dimension.'
    raise ValueError(msg)

  return Ra - Rb


def periodic_displacement(side: Box, dR: Array) -> Array:
  """Wraps displacement vectors into a hypercube.

  Args:
    side: Specification of hypercube size. Either,
      (a) float if all sides have equal length.
      (b) ndarray(spatial_dim) if sides have different lengths.
    dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
  Returns:
    Matrix of wrapped displacements; `ndarray(shape=[..., spatial_dim])`.
  """
  return jnp.mod(dR + side * f32(0.5), side) - f32(0.5) * side


def square_distance(dR: Array) -> Array:
  """Computes square distances.

  Args:
    dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
  Returns:
    Matrix of squared distances; `ndarray(shape=[...])`.
  """
  return jnp.sum(dR ** 2, axis=-1)


def distance(dR: Array) -> Array:
  """Computes distances.

  Args:
    dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
  Returns:
    Matrix of distances; `ndarray(shape=[...])`.
  """
  dr = square_distance(dR)
  return safe_mask(dr > 0, jnp.sqrt, dr)


def periodic_shift(side: Box, R: Array, dR: Array) -> Array:
  """Shifts positions, wrapping them back within a periodic hypercube."""
  return jnp.mod(R + dR, side)


""" Spaces """


def free() -> Space:
  """Free boundary conditions."""
  def displacement_fn(Ra: Array, Rb: Array, perturbation: Optional[Array]=None,
                      **unused_kwargs) -> Array:
    dR = pairwise_displacement(Ra, Rb)
    if perturbation is not None:
      dR = raw_transform(perturbation, dR)
    return dR
  def shift_fn(R: Array, dR: Array, **unused_kwargs) -> Array:
    return R + dR
  return displacement_fn, shift_fn


def periodic(side: Box, wrapped: bool=True) -> Space:
  """Periodic boundary conditions on a hypercube of sidelength side.

  Args:
    side: Either a float or an ndarray of shape [spatial_dimension] specifying
      the size of each side of the periodic box.
    wrapped: A boolean specifying whether or not particle positions are
      remapped back into the box after each step
  Returns:
    `(displacement_fn, shift_fn)` tuple.
  """
  def displacement_fn(Ra: Array, Rb: Array,
                      perturbation: Optional[Array] = None,
                      **unused_kwargs) -> Array:
    if 'box' in unused_kwargs:
      raise UnexpectedBoxException(('`space.periodic` does not accept a box '
                                    'argument. Perhaps you meant to use '
                                    '`space.periodic_general`?'))
    dR = periodic_displacement(side, pairwise_displacement(Ra, Rb))
    if perturbation is not None:
      dR = raw_transform(perturbation, dR)
    return dR
  if wrapped:
    def shift_fn(R: Array, dR: Array, **unused_kwargs) -> Array:
      if 'box' in unused_kwargs:
        raise UnexpectedBoxException(('`space.periodic` does not accept a box '
                                      'argument. Perhaps you meant to use '
                                      '`space.periodic_general`?'))

      return periodic_shift(side, R, dR)
  else:
    def shift_fn(R: Array, dR: Array, **unused_kwargs) -> Array:
      if 'box' in unused_kwargs:
        raise UnexpectedBoxException(('`space.periodic` does not accept a box '
                                      'argument. Perhaps you meant to use '
                                      '`space.periodic_general`?'))
      return R + dR
  return displacement_fn, shift_fn


def periodic_general(box: Box,
                     fractional_coordinates: bool=True,
                     wrapped: bool=True) -> Space:
  """Periodic boundary conditions on a parallelepiped.

  This function defines a simulation on a parallelepiped, :math:`X`, formed by
  applying an affine transformation, :math:`T`, to the unit hypercube
  :math:`U = [0, 1]^d` along with periodic boundary conditions across all
  of the faces.

  Formally, the space is defined such that :math:`X = {Tu : u \in [0, 1]^d}`.

  The affine transformation, :math:`T`, can be specified in a number of different
  ways. For a parallelepiped that is: 1) a cube of side length :math:`L`, the affine
  transformation can simply be a scalar; 2) an orthorhombic unit cell can be
  specified by a vector `[Lx, Ly, Lz]` of lengths for each axis; 3) a general
  triclinic cell can be specified by an upper triangular matrix.

  There are a number of ways to parameterize a simulation on :math:`X`.
  `periodic_general` supports two parametrizations of :math:`X` that can be selected
  using the `fractional_coordinates` keyword argument.

    1) When `fractional_coordinates=True`, particle positions are stored in the
       unit cube, :math:`u\in U`. Here, the displacement function computes the
       displacement between :math:`x, y \in X` as :math:`d_X(x, y) = Td_U(u, v)` where
       :math:`d_U` is the displacement function on the unit cube, :math:`U`, :math:`x = Tu`, and
       :math:`v = Tv` with :math:`u, v \in U`. The derivative of the displacement function
       is defined so that derivatives live in :math:`X` (as opposed to being
       backpropagated to :math:`U`). The shift function, `shift_fn(R, dR)` is defined
       so that :math:`R` is expected to lie in :math:`U` while :math:`dR` should lie in :math:`X`. This
       combination enables code such as `shift_fn(R, force_fn(R))` to work as
       intended.

    2) When `fractional_coordinates=False`, particle positions are stored in
       the parallelepiped :math:`X`. Here, for :math:`x, y \in X`, the displacement function
       is defined as :math:`d_X(x, y) = Td_U(T^{-1}x, T^{-1}y)`. Since there is an
       extra multiplication by :math:`T^{-1}`, this parameterization is typically
       slower than `fractional_coordinates=False`. As in 1), the displacement
       function is defined to compute derivatives in :math:`X`. The shift function
       is defined so that :math:`R` and :math:`dR` should both lie in :math:`X`.

  Example:
  
  .. code-block:: python

     from jax import random
     side_length = 10.0
     disp_frac, shift_frac = periodic_general(side_length,
                                               fractional_coordinates=True)
     disp_real, shift_real = periodic_general(side_length,
                                               fractional_coordinates=False) 

     # Instantiate random positions in both parameterizations.
     R_frac = random.uniform(random.PRNGKey(0), (4, 3))
     R_real = side_length * R_frac

     # Make some shift vectors.
     dR = random.normal(random.PRNGKey(0), (4, 3))

     disp_real(R_real[0], R_real[1]) == disp_frac(R_frac[0], R_frac[1])
     transform(side_length, shift_frac(R_frac, 1.0)) == shift_real(R_real, 1.0)

  It is often desirable to deform a simulation cell either: using a finite
  deformation during a simulation, or using an infinitesimal deformation while
  computing elastic constants. To do this using fractional coordinates, we can
  supply a new affine transformation as `displacement_fn(Ra, Rb, box=new_box)`.
  When using real coordinates, we can specify positions in a space :math:`X` defined
  by an affine transformation :math:`T` and compute displacements in a deformed space
  :math:`X'` defined by an affine transformation :math:`T'`. This is done by writing
  `displacement_fn(Ra, Rb, new_box=new_box)`.

  There are a few caveats when using `periodic_general`. `periodic_general`
  uses the minimum image convention, and so it will fail for potentials whose
  cutoff is longer than the half of the side-length of the box. It will also
  fail to find the correct image when the box is too deformed. We hope to add a
  more robust box for small simulations soon (TODO) along with better error
  checking. In the meantime caution is recommended.

  Args:
    box: A `(spatial_dim, spatial_dim)` affine transformation.
    fractional_coordinates: A boolean specifying whether positions are stored
      in the parallelepiped or the unit cube.
    wrapped: A boolean specifying whether or not particle positions are
      remapped back into the box after each step
  Returns:
    `(displacement_fn, shift_fn)` tuple.
  """
  inv_box = inverse(box)

  def displacement_fn(Ra, Rb, perturbation=None, **kwargs):
    _box, _inv_box = box, inv_box

    if 'box' in kwargs:
      _box = kwargs['box']

      if not fractional_coordinates:
        _inv_box = inverse(_box)

    if 'new_box' in kwargs:
      _box = kwargs['new_box']

    if not fractional_coordinates:
      Ra = transform(_inv_box, Ra)
      Rb = transform(_inv_box, Rb)

    dR = periodic_displacement(f32(1.0), pairwise_displacement(Ra, Rb))
    dR = transform(_box, dR)

    if perturbation is not None:
      dR = raw_transform(perturbation, dR)

    return dR

  def u(R, dR):
    if wrapped:
      return periodic_shift(f32(1.0), R, dR)
    return R + dR

  def shift_fn(R, dR, **kwargs):
    if not fractional_coordinates and not wrapped:
      return R + dR

    _box, _inv_box = box, inv_box
    if 'box' in kwargs:
      _box = kwargs['box']
      _inv_box = inverse(_box)

    if 'new_box' in kwargs:
      _box = kwargs['new_box']

    dR = transform(_inv_box, dR)
    if not fractional_coordinates:
      R = transform(_inv_box, R)

    R = u(R, dR)

    if not fractional_coordinates:
      R = transform(_box, R)
    return R

  return displacement_fn, shift_fn


def metric(displacement: DisplacementFn) -> MetricFn:
  """Takes a displacement function and creates a metric."""
  return lambda Ra, Rb, **kwargs: distance(displacement(Ra, Rb, **kwargs))


def map_product(metric_or_displacement: DisplacementOrMetricFn
                ) -> DisplacementOrMetricFn:
  """Vectorizes a metric or displacement function over all pairs."""
  return vmap(vmap(metric_or_displacement, (0, None), 0), (None, 0), 0)


def map_bond(metric_or_displacement: DisplacementOrMetricFn
             ) -> DisplacementOrMetricFn:
  """Vectorizes a metric or displacement function over bonds."""
  return vmap(metric_or_displacement, (0, 0), 0)


def map_neighbor(metric_or_displacement: DisplacementOrMetricFn
                 ) -> DisplacementOrMetricFn:
  """Vectorizes a metric or displacement function over neighborhoods."""
  def wrapped_fn(Ra, Rb, **kwargs):
    return vmap(vmap(metric_or_displacement, (None, 0)))(-Ra, -Rb, **kwargs)
  return wrapped_fn


def canonicalize_displacement_or_metric(displacement_or_metric):
  """Checks whether or not a displacement or metric was provided."""
  for dim in range(1, 4):
    try:
      R = ShapedArray((dim,), f32)
      dR_or_dr = eval_shape(displacement_or_metric, R, R, t=0)
      if len(dR_or_dr.shape) == 0:
        return displacement_or_metric
      else:
        return metric(displacement_or_metric)
    except TypeError:
      continue
    except ValueError:
      continue
  raise ValueError(
    'Canonicalize displacement not implemented for spatial dimension larger'
    'than 4.')




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for defining dataclasses that can be used with jax transformations.

This code was copied and adapted from https://github.com/google/flax/struct.py.

Accessed on 04/29/2020.
"""

import dataclasses
import jax


def dataclass(clz):
  """Create a class which can be passed to functional transformations.

  Jax transformations such as `jax.jit` and `jax.grad` require objects that are
  immutable and can be mapped over using the `jax.tree_util` methods.

  The `dataclass` decorator makes it easy to define custom classes that can be
  passed safely to Jax.

  Args:
    clz: the class that will be transformed by the decorator.
  Returns:
    The new class.
  """
  data_clz = dataclasses.dataclass(frozen=True)(clz)
  meta_fields = []
  data_fields = []
  for name, field_info in data_clz.__dataclass_fields__.items():
    is_static = field_info.metadata.get('static', False)
    if is_static:
      meta_fields.append(name)
    else:
      data_fields.append(name)

  def iterate_clz(x):
    meta = tuple(getattr(x, name) for name in meta_fields)
    data = tuple(getattr(x, name) for name in data_fields)
    return data, meta

  def clz_from_iterable(meta, data):
    meta_args = tuple(zip(meta_fields, meta))
    data_args = tuple(zip(data_fields, data))
    kwargs = dict(meta_args + data_args)
    return data_clz(**kwargs)

  jax.tree_util.register_pytree_node(data_clz,
                                     iterate_clz,
                                     clz_from_iterable)

  return data_clz


def static_field():
  return dataclasses.field(metadata={'static': True})

replace = dataclasses.replace
asdict = dataclasses.asdict
astuple = dataclasses.astuple
is_dataclass = dataclasses.is_dataclass
fields = dataclasses.fields
def unpack(dc) -> tuple:
    return tuple(getattr(dc, field.name) for field in dataclasses.fields(dc))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to transform functions on individual tuples of particles to sets."""

from absl import logging

from functools import reduce, partial
from collections import namedtuple

from enum import Enum

from typing import Any, Callable, Optional, Dict, Tuple, Generator, Union

import math
from operator import mul

import numpy as onp

from jax import lax
from jax import ops
from jax import jit, vmap, eval_shape
from jax.abstract_arrays import ShapedArray
from jax.interpreters import partial_eval as pe
import jax.numpy as jnp

from jax_md import space
from jax_md import dataclasses
from jax_md import util

import jraph


# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

i32 = util.i32
i64 = util.i64

Box = space.Box
DisplacementOrMetricFn = space.DisplacementOrMetricFn
MetricFn = space.MetricFn
MaskFn = Callable[[Array], Array]

# Cell List


@dataclasses.dataclass
class CellList:
  """Stores the spatial partition of a system into a cell list.

  See :meth:`cell_list` for details on the construction / specification.
  Cell list buffers all have a common shape, S, where
    * `S = [cell_count_x, cell_count_y, cell_capacity]`
    * `S = [cell_count_x, cell_count_y, cell_count_z, cell_capacity]`
  in two- and three-dimensions respectively. It is assumed that each cell has
  the same capacity.

  Attributes:
    position_buffer: An ndarray of floating point positions with shape
      `S + [spatial_dimension]`.
    id_buffer: An ndarray of int32 particle ids of shape `S`. Note that empty
      slots are specified by `id = N` where `N` is the number of particles in the
      system.
    kwarg_buffers: A dictionary of ndarrays of shape `S + [...]`. This contains
      side data placed into the cell list.
    did_buffer_overflow: A boolean specifying whether or not the cell list
      exceeded the maximum allocated capacity.
    cell_capacity: An integer specifying the maximum capacity of each cell in
      the cell list.
    update_fn: A function that updates the cell list at a fixed capacity.
  """
  position_buffer: Array
  id_buffer: Array
  kwarg_buffers: Dict[str, Array]

  did_buffer_overflow: Array

  cell_capacity: int = dataclasses.static_field()

  update_fn: Callable[..., 'CellList'] = \
      dataclasses.static_field()

  def update(self, position: Array, **kwargs) -> 'CellList':
    cl_data = (self.cell_capacity, self.did_buffer_overflow, self.update_fn)
    return self.update_fn(position, cl_data, **kwargs)


@dataclasses.dataclass
class CellListFns:
  allocate: Callable[..., CellList] = dataclasses.static_field()
  update: Callable[[Array, Union[CellList, int]],
                    CellList] = dataclasses.static_field()

  def __iter__(self):
    return iter((self.allocate, self.update))


def _cell_dimensions(spatial_dimension: int,
                     box_size: Box,
                     minimum_cell_size: float) -> Tuple[Box, Array, Array, int]:
  """Compute the number of cells-per-side and total number of cells in a box."""
  if isinstance(box_size, int) or isinstance(box_size, float):
    box_size = float(box_size)

  # NOTE(schsam): Should we auto-cast based on box_size? I can't imagine a case
  # in which the box_size would not be accurately represented by an f32.
  if (isinstance(box_size, onp.ndarray) and
      (box_size.dtype == i32 or box_size.dtype == i64)):
    box_size = float(box_size)

  cells_per_side = onp.floor(box_size / minimum_cell_size)
  cell_size = box_size / cells_per_side
  cells_per_side = onp.array(cells_per_side, dtype=i32)

  if isinstance(box_size, onp.ndarray):
    if box_size.ndim == 1 or box_size.ndim == 2:
      assert box_size.size == spatial_dimension
      flat_cells_per_side = onp.reshape(cells_per_side, (-1,))
      for cells in flat_cells_per_side:
        if cells < 3:
          msg = ('Box must be at least 3x the size of the grid spacing in each '
                 'dimension.')
          raise ValueError(msg)
      cell_count = reduce(mul, flat_cells_per_side, 1)
    elif box_size.ndim == 0:
      cell_count = cells_per_side ** spatial_dimension
    else:
      raise ValueError('Box must either be a scalar or a vector.')
  else:
    cell_count = cells_per_side ** spatial_dimension

  return box_size, cell_size, cells_per_side, int(cell_count)


def count_cell_filling(position: Array,
                       box_size: Box,
                       minimum_cell_size: float) -> Array:
  """Counts the number of particles per-cell in a spatial partition."""
  dim = int(position.shape[1])
  box_size, cell_size, cells_per_side, cell_count = \
      _cell_dimensions(dim, box_size, minimum_cell_size)

  hash_multipliers = _compute_hash_constants(dim, cells_per_side)

  particle_index = jnp.array(position / cell_size, dtype=i32)
  particle_hash = jnp.sum(particle_index * hash_multipliers, axis=1)

  filling = ops.segment_sum(jnp.ones_like(particle_hash),
                            particle_hash,
                            cell_count)
  return filling


def _compute_hash_constants(spatial_dimension: int,
                            cells_per_side: Array) -> Array:
  if cells_per_side.size == 1:
    return jnp.array([[cells_per_side ** d for d in range(spatial_dimension)]],
                     dtype=i32)
  elif cells_per_side.size == spatial_dimension:
    one = jnp.array([[1]], dtype=i32)
    cells_per_side = jnp.concatenate((one, cells_per_side[:, :-1]), axis=1)
    return jnp.array(jnp.cumprod(cells_per_side), dtype=i32)
  else:
    raise ValueError()


def _neighboring_cells(dimension: int) -> Generator[onp.ndarray, None, None]:
  for dindex in onp.ndindex(*([3] * dimension)):
    yield onp.array(dindex, dtype=i32) - 1


def _estimate_cell_capacity(position: Array,
                            box_size: Box,
                            cell_size: float,
                            buffer_size_multiplier: float) -> int:
  cell_capacity = onp.max(count_cell_filling(position, box_size, cell_size))
  return int(cell_capacity * buffer_size_multiplier)


def _shift_array(arr: Array, dindex: Array) -> Array:
  if len(dindex) == 2:
    dx, dy = dindex
    dz = 0
  elif len(dindex) == 3:
    dx, dy, dz = dindex

  if dx < 0:
    arr = jnp.concatenate((arr[1:], arr[:1]))
  elif dx > 0:
    arr = jnp.concatenate((arr[-1:], arr[:-1]))

  if dy < 0:
    arr = jnp.concatenate((arr[:, 1:], arr[:, :1]), axis=1)
  elif dy > 0:
    arr = jnp.concatenate((arr[:, -1:], arr[:, :-1]), axis=1)

  if dz < 0:
    arr = jnp.concatenate((arr[:, :, 1:], arr[:, :, :1]), axis=2)
  elif dz > 0:
    arr = jnp.concatenate((arr[:, :, -1:], arr[:, :, :-1]), axis=2)

  return arr


def _unflatten_cell_buffer(arr: Array,
                           cells_per_side: Array,
                           dim: int) -> Array:
  if (isinstance(cells_per_side, int) or
      isinstance(cells_per_side, float) or
      (util.is_array(cells_per_side) and not cells_per_side.shape)):
    cells_per_side = (int(cells_per_side),) * dim
  elif util.is_array(cells_per_side) and len(cells_per_side.shape) == 1:
    cells_per_side = tuple([int(x) for x in cells_per_side[::-1]])
  elif util.is_array(cells_per_side) and len(cells_per_side.shape) == 2:
    cells_per_side = tuple([int(x) for x in cells_per_side[0][::-1]])
  else:
    raise ValueError()
  return jnp.reshape(arr, cells_per_side + (-1,) + arr.shape[1:])


def cell_list(box_size: Box,
              minimum_cell_size: float,
              buffer_size_multiplier: float = 1.25
              ) -> CellListFns:
  r"""Returns a function that partitions point data spatially.

  Given a set of points :math:`\{x_i \in R^d\}` with associated data :math:`\{k_i \in R^m\}` 
  it is often useful to partition the points / data spatially. A simple partitioning
  that can be implemented efficiently within XLA is a dense partition into a
  uniform grid called a cell list.

  Since XLA requires that shapes be statically specified inside of a JIT block,
  the cell list code can operate in two modes: allocation and update.

  Allocation creates a new cell list that uses a set of input positions to
  estimate the capacity of the cell list. This capacity can be adjusted by
  setting the `buffer_size_multiplier` or setting the `extra_capacity`.
  Allocation cannot be JIT.

  Updating takes a previously allocated cell list and places a new set of
  particles in the cells. Updating cannot resize the cell list and is therefore
  compatible with JIT. However, if the configuration has changed substantially
  it is possible that the existing cell list won't be large enough to
  accommodate all of the particles. In this case the `did_buffer_overflow` bit
  will be set to True.

  Args:
    box_size: A float or an ndarray of shape `[spatial_dimension]` specifying the
      size of the system. Note, this code is written for the case where the
      boundaries are periodic. If this is not the case, then the current code
      will be slightly less efficient.
    minimum_cell_size: A float specifying the minimum side length of each cell.
      Cells are enlarged so that they exactly fill the box.
    buffer_size_multiplier: A floating point multiplier that multiplies the
      estimated cell capacity to allow for fluctuations in the maximum cell
      occupancy.
  Returns:
    A `CellListFns` object that contains two methods, one to allocate the cell
    list and one to update the cell list. The update function can be called
    with either a cell list from which the capacity can be inferred or with
    an explicit integer denoting the capacity. Note that an existing cell list
    can also be updated by calling `cell_list.update(position)`.
  """

  if util.is_array(box_size):
    box_size = onp.array(box_size)
    if len(box_size.shape) == 1:
      box_size = jnp.reshape(box_size, (1, -1))

  if util.is_array(minimum_cell_size):
    minimum_cell_size = onp.array(minimum_cell_size)

  def cell_list_fn(position: Array,
                   capacity_overflow_update: Optional[
                       Tuple[int, bool, Callable[..., CellList]]] = None,
                   extra_capacity: int = 0, **kwargs) -> CellList:
    N = position.shape[0]
    dim = position.shape[1]

    if dim != 2 and dim != 3:
      # NOTE(schsam): Do we want to check this in compute_fn as well?
      raise ValueError(
          f'Cell list spatial dimension must be 2 or 3. Found {dim}.')

    _, cell_size, cells_per_side, cell_count = \
        _cell_dimensions(dim, box_size, minimum_cell_size)

    if capacity_overflow_update is None:
      cell_capacity = _estimate_cell_capacity(position, box_size, cell_size,
                                              buffer_size_multiplier)
      cell_capacity += extra_capacity
      overflow = False
      update_fn = cell_list_fn
    else:
      cell_capacity, overflow, update_fn = capacity_overflow_update

    hash_multipliers = _compute_hash_constants(dim, cells_per_side)

    # Create cell list data.
    particle_id = lax.iota(i32, N)
    # NOTE(schsam): We use the convention that particles that are successfully,
    # copied have their true id whereas particles empty slots have id = N.
    # Then when we copy data back from the grid, copy it to an array of shape
    # [N + 1, output_dimension] and then truncate it to an array of shape
    # [N, output_dimension] which ignores the empty slots.
    cell_position = jnp.zeros((cell_count * cell_capacity, dim),
                              dtype=position.dtype)
    cell_id = N * jnp.ones((cell_count * cell_capacity, 1), dtype=i32)

    # It might be worth adding an occupied mask. However, that will involve
    # more compute since often we will do a mask for species that will include
    # an occupancy test. It seems easier to design around this empty_data_value
    # for now and revisit the issue if it comes up later.
    empty_kwarg_value = 10 ** 5
    cell_kwargs = {}
    #  pytype: disable=attribute-error
    for k, v in kwargs.items():
      if not util.is_array(v):
        raise ValueError((f'Data must be specified as an ndarray. Found "{k}" '
                          f'with type {type(v)}.'))
      if v.shape[0] != position.shape[0]:
        raise ValueError(('Data must be specified per-particle (an ndarray '
                          f'with shape ({N}, ...)). Found "{k}" with '
                          f'shape {v.shape}.'))
      kwarg_shape = v.shape[1:] if v.ndim > 1 else (1,)
      cell_kwargs[k] = empty_kwarg_value * jnp.ones(
          (cell_count * cell_capacity,) + kwarg_shape, v.dtype)
    #  pytype: enable=attribute-error
    indices = jnp.array(position / cell_size, dtype=i32)
    hashes = jnp.sum(indices * hash_multipliers, axis=1)

    # Copy the particle data into the grid. Here we use a trick to allow us to
    # copy into all cells simultaneously using a single lax.scatter call. To do
    # this we first sort particles by their cell hash. We then assign each
    # particle to have a cell id = hash * cell_capacity + grid_id where
    # grid_id is a flat list that repeats 0, .., cell_capacity. So long as
    # there are fewer than cell_capacity particles per cell, each particle is
    # guaranteed to get a cell id that is unique.
    sort_map = jnp.argsort(hashes)
    sorted_position = position[sort_map]
    sorted_hash = hashes[sort_map]
    sorted_id = particle_id[sort_map]

    sorted_kwargs = {}
    for k, v in kwargs.items():
      sorted_kwargs[k] = v[sort_map]

    sorted_cell_id = jnp.mod(lax.iota(i32, N), cell_capacity)
    sorted_cell_id = sorted_hash * cell_capacity + sorted_cell_id

    cell_position = cell_position.at[sorted_cell_id].set(sorted_position)
    sorted_id = jnp.reshape(sorted_id, (N, 1))
    cell_id = cell_id.at[sorted_cell_id].set(sorted_id)
    cell_position = _unflatten_cell_buffer(cell_position, cells_per_side, dim)
    cell_id = _unflatten_cell_buffer(cell_id, cells_per_side, dim)

    for k, v in sorted_kwargs.items():
      if v.ndim == 1:
        v = jnp.reshape(v, v.shape + (1,))
      cell_kwargs[k] = cell_kwargs[k].at[sorted_cell_id].set(v)
      cell_kwargs[k] = _unflatten_cell_buffer(
          cell_kwargs[k], cells_per_side, dim)

    occupancy = ops.segment_sum(jnp.ones_like(hashes), hashes, cell_count)
    max_occupancy = jnp.max(occupancy)
    overflow = overflow | (max_occupancy >= cell_capacity)

    return CellList(cell_position, cell_id, cell_kwargs,
                    overflow, cell_capacity, update_fn)  # pytype: disable=wrong-arg-count

  def allocate_fn(position: Array, extra_capacity: int = 0, **kwargs
                  ) -> CellList:
    return cell_list_fn(position, extra_capacity=extra_capacity, **kwargs)

  def update_fn(position: Array, cl_or_capacity: Union[CellList, int], **kwargs
                ) -> CellList:
    if isinstance(cl_or_capacity, int):
      capacity = int(cl_or_capacity)
      return cell_list_fn(position, (capacity, False, cell_list_fn), **kwargs)
    cl = cl_or_capacity
    cl_data = (cl.cell_capacity, cl.did_buffer_overflow, cl.update_fn)
    return cell_list_fn(position, cl_data, **kwargs)

  return CellListFns(allocate_fn, update_fn)  # pytype: disable=wrong-arg-count


def _displacement_or_metric_to_metric_sq(
    displacement_or_metric: DisplacementOrMetricFn) -> MetricFn:
  """Checks whether or not a displacement or metric was provided."""
  for dim in range(1, 4):
    try:
      R = ShapedArray((dim,), f32)
      dR_or_dr = eval_shape(displacement_or_metric, R, R, t=0)
      if len(dR_or_dr.shape) == 0:
        return lambda Ra, Rb, **kwargs: \
          displacement_or_metric(Ra, Rb, **kwargs) ** 2
      else:
        return lambda Ra, Rb, **kwargs: space.square_distance(
          displacement_or_metric(Ra, Rb, **kwargs))
    except TypeError:
      continue
    except ValueError:
      continue
  raise ValueError(
    'Canonicalize displacement not implemented for spatial dimension larger'
    'than 4.')


class NeighborListFormat(Enum):
  """An enum listing the different neighbor list formats.

  Attributes:
    Dense: A dense neighbor list where the ids are a square matrix
      of shape `(N, max_neighbors_per_atom)`. Here the capacity of the neighbor
      list must scale with the highest connectivity neighbor.
    Sparse: A sparse neighbor list where the ids are a rectangular
      matrix of shape `(2, max_neighbors)` specifying the start / end particle
      of each neighbor pair.
    OrderedSparse: A sparse neighbor list whose format is the same as `Sparse`
      where only bonds with i < j are included.
  """
  Dense = 0
  Sparse = 1
  OrderedSparse = 2


def is_sparse(fmt: NeighborListFormat) -> bool:
  return (fmt is NeighborListFormat.Sparse or
          fmt is NeighborListFormat.OrderedSparse)


def is_format_valid(fmt: NeighborListFormat):
  if fmt not in list(NeighborListFormat):
    raise ValueError((
        'Neighbor list format must be a member of NeighborListFormat'
        f' found {fmt}.'))


@dataclasses.dataclass
class NeighborList(object):
  """A struct containing the state of a Neighbor List.

  Attributes:
    idx: For an N particle system this is an `[N, max_occupancy]` array of
      integers such that `idx[i, j]` is the j-th neighbor of particle i.
    reference_position: The positions of particles when the neighbor list was
      constructed. This is used to decide whether the neighbor list ought to be
      updated.
    did_buffer_overflow: A boolean that starts out False. If there are ever
      more neighbors than max_neighbors this is set to true to indicate that
      there was a buffer overflow. If this happens, it means that the results
      of the simulation will be incorrect and the simulation needs to be rerun
      using a larger buffer.
    cell_list_capacity: An optional integer specifying the capacity of the cell
      list used as an intermediate step in the creation of the neighbor list.
    max_occupancy: A static integer specifying the maximum size of the
      neighbor list. Changing this will invoke a recompilation.
    format: A NeighborListFormat enum specifying the format of the neighbor
      list.
    update_fn: A static python function used to update the neighbor list.
  """
  idx: Array

  reference_position: Array

  did_buffer_overflow: Array

  cell_list_capacity: Optional[int] = dataclasses.static_field()

  max_occupancy: int = dataclasses.static_field()

  format: NeighborListFormat = dataclasses.static_field()
  update_fn: Callable[[Array, 'NeighborList'],
                      'NeighborList'] = dataclasses.static_field()

  def update(self, position: Array, **kwargs) -> 'NeighborList':
    return self.update_fn(position, self, **kwargs)


@dataclasses.dataclass
class NeighborListFns:
  """A struct containing functions to allocate and update neighbor lists.

  Attributes:
    allocate: A function to allocate a new neighbor list. This function cannot
      be compiled, since it uses the values of positions to infer the shapes.
    update: A function to update a neighbor list given a new set of positions
      and a previously allocated neighbor list.
  """
  allocate: Callable[..., NeighborList] = dataclasses.static_field()
  update: Callable[[Array, NeighborList],
                   NeighborList] = dataclasses.static_field()

  def __call__(self,
               position: Array,
               neighbors: Optional[NeighborList] = None,
               extra_capacity: int = 0,
               **kwargs) -> NeighborList:
    """A function for backward compatibility with previous neighbor lists.

    Args:
      position: An `(N, dim)` array of particle positions.
      neighbors: An optional neighbor list object. If it is provided then
        the function updates the neighbor list, otherwise it allocates a new
        neighbor list.
      extra_capacity: Extra capacity to add if allocating the neighbor list.
    Returns:
      A neighbor list object.
    """
    logging.warning('Using a deprecated code path to create / update neighbor '
                    'lists. It will be removed in a later version of JAX MD. '
                    'Using `neighbor_fn.allocate` and `neighbor_fn.update` '
                    'is preferred.')
    if neighbors is None:
      return self.allocate(position, extra_capacity, **kwargs)
    return self.update(position, neighbors, **kwargs)

  def __iter__(self):
    return iter((self.allocate, self.update))


NeighborFn = Callable[[Array, Optional[NeighborList], Optional[int]],
                      NeighborList]


def neighbor_list(displacement_or_metric: DisplacementOrMetricFn,
                  box_size: Box,
                  r_cutoff: float,
                  dr_threshold: float,
                  capacity_multiplier: float = 1.25,
                  disable_cell_list: bool = False,
                  mask_self: bool = True,
                  custom_mask_function: Optional[MaskFn] = None,
                  fractional_coordinates: bool = False,
                  format: NeighborListFormat = NeighborListFormat.Dense,
                  **static_kwargs) -> NeighborFn:
  """Returns a function that builds a list neighbors for collections of points.

  Neighbor lists must balance the need to be jit compatible with the fact that
  under a jit the maximum number of neighbors cannot change (owing to static
  shape requirements). To deal with this, our `neighbor_list` returns a
  `NeighborListFns` object that contains two functions: 1)
  `neighbor_fn.allocate` create a new neighbor list and 2) `neighbor_fn.update`
  updates an existing neighbor list. Neighbor lists themselves additionally
  have a convenience `update` member function.

  Note that allocation of a new neighbor list cannot be jit compiled since it
  uses the positions to infer the maximum number of neighbors (along with
  additional space specified by the `capacity_multiplier`). Updating the
  neighbor list can be jit compiled; if the neighbor list capacity is not
  sufficient to store all the neighbors, the `did_buffer_overflow` bit
  will be set to `True` and a new neighbor list will need to be reallocated.

  Here is a typical example of a simulation loop with neighbor lists:

  .. code-block:: python

     init_fn, apply_fn = simulate.nve(energy_fn, shift, 1e-3)
     exact_init_fn, exact_apply_fn = simulate.nve(exact_energy_fn, shift, 1e-3)

     nbrs = neighbor_fn.allocate(R)
     state = init_fn(random.PRNGKey(0), R, neighbor_idx=nbrs.idx)

     def body_fn(i, state):
       state, nbrs = state
       nbrs = nbrs.update(state.position)
       state = apply_fn(state, neighbor_idx=nbrs.idx)
       return state, nbrs

     step = 0
     for _ in range(20):
       new_state, nbrs = lax.fori_loop(0, 100, body_fn, (state, nbrs))
       if nbrs.did_buffer_overflow:
         nbrs = neighbor_fn.allocate(state.position)
       else:
         state = new_state
         step += 1

  Args:
    displacement: A function `d(R_a, R_b)` that computes the displacement
      between pairs of points.
    box_size: Either a float specifying the size of the box or an array of
      shape `[spatial_dim]` specifying the box size in each spatial dimension.
    r_cutoff: A scalar specifying the neighborhood radius.
    dr_threshold: A scalar specifying the maximum distance particles can move
      before rebuilding the neighbor list.
    capacity_multiplier: A floating point scalar specifying the fractional
      increase in maximum neighborhood occupancy we allocate compared with the
      maximum in the example positions.
    disable_cell_list: An optional boolean. If set to `True` then the neighbor
      list is constructed using only distances. This can be useful for
      debugging but should generally be left as `False`.
    mask_self: An optional boolean. Determines whether points can consider
      themselves to be their own neighbors.
    custom_mask_function: An optional function. Takes the neighbor array
      and masks selected elements. Note: The input array to the function is
      `(n_particles, m)` where the index of particle 1 is in index in the first
      dimension of the array, the index of particle 2 is given by the value in
      the array
    fractional_coordinates: An optional boolean. Specifies whether positions
      will be supplied in fractional coordinates in the unit cube, :math:`[0, 1]^d`.
      If this is set to True then the `box_size` will be set to `1.0` and the
      cell size used in the cell list will be set to `cutoff / box_size`.
    format: The format of the neighbor list; see the :meth:`NeighborListFormat` enum
      for details about the different choices for formats. Defaults to `Dense`.
    **static_kwargs: kwargs that get threaded through the calculation of
      example positions.
  Returns:
    A NeighborListFns object that contains a method to allocate a new neighbor
    list and a method to update an existing neighbor list.
  """
  is_format_valid(format)
  box_size = lax.stop_gradient(box_size)
  r_cutoff = lax.stop_gradient(r_cutoff)
  dr_threshold = lax.stop_gradient(dr_threshold)

  box_size = f32(box_size)

  cutoff = r_cutoff + dr_threshold
  cutoff_sq = cutoff ** 2
  threshold_sq = (dr_threshold / f32(2)) ** 2
  metric_sq = _displacement_or_metric_to_metric_sq(displacement_or_metric)

  cell_size = cutoff
  if fractional_coordinates:
    cell_size = cutoff / box_size
    box_size = (f32(box_size) if onp.isscalar(box_size)
                else onp.ones_like(box_size, f32))

  use_cell_list = jnp.all(cell_size < box_size / 3.) and not disable_cell_list
  if use_cell_list:
    cl_fn = cell_list(box_size, cell_size, capacity_multiplier)

  @jit
  def candidate_fn(position: Array) -> Array:
    candidates = jnp.arange(position.shape[0])
    return jnp.broadcast_to(candidates[None, :],
                            (position.shape[0], position.shape[0]))

  @jit
  def cell_list_candidate_fn(cl: CellList, position: Array) -> Array:
    N, dim = position.shape

    idx = cl.id_buffer

    cell_idx = [idx]

    for dindex in _neighboring_cells(dim):
      if onp.all(dindex == 0):
        continue
      cell_idx += [_shift_array(idx, dindex)]

    cell_idx = jnp.concatenate(cell_idx, axis=-2)
    cell_idx = cell_idx[..., jnp.newaxis, :, :]
    cell_idx = jnp.broadcast_to(cell_idx, idx.shape[:-1] + cell_idx.shape[-2:])

    def copy_values_from_cell(value, cell_value, cell_id):
      scatter_indices = jnp.reshape(cell_id, (-1,))
      cell_value = jnp.reshape(cell_value, (-1,) + cell_value.shape[-2:])
      return value.at[scatter_indices].set(cell_value)

    neighbor_idx = jnp.zeros((N + 1,) + cell_idx.shape[-2:], i32)
    neighbor_idx = copy_values_from_cell(neighbor_idx, cell_idx, idx)
    return neighbor_idx[:-1, :, 0]

  @jit
  def mask_self_fn(idx: Array) -> Array:
    self_mask = idx == jnp.reshape(jnp.arange(idx.shape[0], dtype=i32), (idx.shape[0], 1))
    return jnp.where(self_mask, idx.shape[0], idx)

  @jit
  def prune_neighbor_list_dense(position: Array, idx: Array, **kwargs) -> Array:
    d = partial(metric_sq, **kwargs)
    d = space.map_neighbor(d)

    N = position.shape[0]
    neigh_position = position[idx]
    dR = d(position, neigh_position)

    mask = (dR < cutoff_sq) & (idx < N)
    out_idx = N * jnp.ones(idx.shape, i32)

    cumsum = jnp.cumsum(mask, axis=1)
    index = jnp.where(mask, cumsum - 1, idx.shape[1] - 1)
    p_index = jnp.arange(idx.shape[0])[:, None]
    out_idx = out_idx.at[p_index, index].set(idx)
    max_occupancy = jnp.max(cumsum[:, -1])

    return out_idx, max_occupancy

  @jit
  def prune_neighbor_list_sparse(position: Array, idx: Array, **kwargs
                                 ) -> Array:
    d = partial(metric_sq, **kwargs)
    d = space.map_bond(d)

    N = position.shape[0]
    sender_idx = jnp.broadcast_to(jnp.arange(N)[:, None], idx.shape)

    sender_idx = jnp.reshape(sender_idx, (-1,))
    receiver_idx = jnp.reshape(idx, (-1,))
    dR = d(position[sender_idx], position[receiver_idx])

    mask = (dR < cutoff_sq) & (receiver_idx < N)
    if format is NeighborListFormat.OrderedSparse:
      mask = mask & (receiver_idx < sender_idx)

    out_idx = N * jnp.ones(receiver_idx.shape, i32)

    cumsum = jnp.cumsum(mask)
    index = jnp.where(mask, cumsum - 1, len(receiver_idx) - 1)
    receiver_idx = out_idx.at[index].set(receiver_idx)
    sender_idx = out_idx.at[index].set(sender_idx)
    max_occupancy = cumsum[-1]

    return jnp.stack((receiver_idx, sender_idx)), max_occupancy

  def neighbor_list_fn(position: Array,
                       neighbors: Optional[NeighborList] = None,
                       extra_capacity: int = 0,
                       **kwargs) -> NeighborList:
    nbrs = neighbors
    def neighbor_fn(position_and_overflow, max_occupancy=None):
      position, overflow = position_and_overflow
      N = position.shape[0]

      if use_cell_list:
        if neighbors is None:
          cl = cl_fn.allocate(position, extra_capacity=extra_capacity)
        else:
          cl = cl_fn.update(position, neighbors.cell_list_capacity)
        overflow = overflow | cl.did_buffer_overflow
        idx = cell_list_candidate_fn(cl, position)
        cl_capacity = cl.cell_capacity
      else:
        cl_capacity = None
        idx = candidate_fn(position)

      if mask_self:
        idx = mask_self_fn(idx)
      if custom_mask_function is not None:
        idx = custom_mask_function(idx)

      if is_sparse(format):
        idx, occupancy = prune_neighbor_list_sparse(position, idx, **kwargs)
      else:
        idx, occupancy = prune_neighbor_list_dense(position, idx, **kwargs)

      if max_occupancy is None:
        _extra_capacity = (extra_capacity if not is_sparse(format)
                           else N * extra_capacity)
        max_occupancy = int(occupancy * capacity_multiplier + _extra_capacity)
        if max_occupancy > idx.shape[-1]:
          max_occupancy = idx.shape[-1]
        if not is_sparse(format):
          capacity_limit = N - 1 if mask_self else N
        elif format is NeighborListFormat.Sparse:
          capacity_limit = N * (N - 1) if mask_self else N**2
        else:
          capacity_limit = N * (N - 1) // 2
        if max_occupancy > capacity_limit:
          max_occupancy = capacity_limit
      idx = idx[:, :max_occupancy]
      update_fn = (neighbor_list_fn if neighbors is None else
                   neighbors.update_fn)
      return NeighborList(
          idx,
          position,
          overflow | (occupancy > max_occupancy),
          cl_capacity,
          max_occupancy,
          format,
          update_fn)  # pytype: disable=wrong-arg-count

    if nbrs is None:
      return neighbor_fn((position, False))

    neighbor_fn = partial(neighbor_fn, max_occupancy=nbrs.max_occupancy)

    d = partial(metric_sq, **kwargs)
    d = vmap(d)
    return lax.cond(
        jnp.any(d(position, nbrs.reference_position) > threshold_sq),
        (position, nbrs.did_buffer_overflow), neighbor_fn,
        nbrs, lambda x: x)

  def allocate_fn(position: Array, extra_capacity: int = 0, **kwargs
                  ) -> NeighborList:
    return neighbor_list_fn(position, extra_capacity=extra_capacity, **kwargs)

  def update_fn(position: Array, neighbors: NeighborList, **kwargs
                ) -> NeighborList:
    return neighbor_list_fn(position, neighbors, **kwargs)

  return NeighborListFns(allocate_fn, update_fn)  # pytype: disable=wrong-arg-count


def neighbor_list_mask(neighbor: NeighborList, mask_self: bool = False
                       ) -> Array:
  """Compute a mask for neighbor list."""
  if is_sparse(neighbor.format):
    mask = neighbor.idx[0] < len(neighbor.reference_position)
    if mask_self:
      mask = mask & (neighbor.idx[0] != neighbor.idx[1])
    return mask

  mask = neighbor.idx < len(neighbor.idx)
  if mask_self:
    N = len(neighbor.reference_position)
    self_mask = neighbor.idx != jnp.reshape(jnp.arange(N, dtype=i32), (N, 1))
    mask = mask & self_mask
  return mask


def to_jraph(neighbor: NeighborList, mask: Array = None) -> jraph.GraphsTuple:
  """Convert a sparse neighbor list to a `jraph.GraphsTuple`.

  As in jraph, padding here is accomplished by adding a ficticious graph with a
  single node.

  Args:
    neighbor: A neighbor list that we will convert to the jraph format. Must be
      sparse.
    mask: An optional mask on the edges.

  Returns:
    A `jraph.GraphsTuple` that contains the topology of the neighbor list.
  """
  if not is_sparse(neighbor.format):
    raise ValueError('Cannot convert a dense neighbor list to jraph format. '
                     'Please use either NeighborListFormat.Sparse or '
                     'NeighborListFormat.OrderedSparse.')

  receivers, senders = neighbor.idx
  N = len(neighbor.reference_position)

  _mask = neighbor_list_mask(neighbor)

  if mask is not None:
    _mask = _mask & mask
    cumsum = jnp.cumsum(_mask)
    index = jnp.where(_mask, cumsum - 1, len(receivers))
    ordered = N * jnp.ones((len(receivers) + 1,), i32)
    receivers = ordered.at[index].set(receivers)[:-1]
    senders = ordered.at[index].set(senders)[:-1]
    mask = receivers < N

  return jraph.GraphsTuple(
      nodes=None,
      edges=None,
      receivers=receivers,
      senders=senders,
      globals=None,
      n_node=jnp.array([N, 1]),
      n_edge=jnp.array([jnp.sum(_mask), jnp.sum(~_mask)]),
  )


def to_dense(neighbor: NeighborList) -> Array:
  """Converts a sparse neighbor list to dense ids. Cannot be JIT."""
  if neighbor.format is not Sparse:
    raise ValueError('Can only convert sparse neighbor lists to dense ones.')

  receivers, senders = neighbor.idx
  mask = neighbor_list_mask(neighbor)

  receivers = receivers[mask]
  senders = senders[mask]

  N = len(neighbor.reference_position)
  count = ops.segment_sum(jnp.ones(len(receivers), i32), receivers, N)
  max_count = jnp.max(count)
  offset = jnp.tile(jnp.arange(max_count), N)[:len(senders)]
  hashes = senders * max_count + offset
  dense_idx = N * jnp.ones((N * max_count,), i32)
  dense_idx = dense_idx.at[hashes].set(receivers).reshape((N, max_count))
  return dense_idx


Dense = NeighborListFormat.Dense
Sparse = NeighborListFormat.Sparse
OrderedSparse = NeighborListFormat.OrderedSparse



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to simulate systems in various statistical ensembles.

  This file contains a number of different methods that can be used to
  simulate systems in a variety of ensembles.

  In general, simulation code follows the same overall structure as optimizers
  in JAX. Simulations are tuples of two functions:

    init_fn: 
      Function that initializes the  state of a system. Should take
      positions as an ndarray of shape `[n, output_dimension]`. Returns a state
      which will be a namedtuple.
    apply_fn: 
      Function that takes a state and produces a new state after one
      step of optimization.

  One question that we need to think about is whether the simulations should
  also return a function that computes the invariant for that ensemble. This
  can be used for testing purposes, but is not often used otherwise.
"""

from collections import namedtuple

from typing import Callable, TypeVar, Union, Tuple, Dict, Optional

from jax import grad
from jax import random
import jax.numpy as jnp
from jax import lax

from jax_md import quantity
from jax_md import util
from jax_md import space
from jax_md import dataclasses
from jax_md import partition
from jax_md import smap

static_cast = util.static_cast


# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

Box = space.Box

ShiftFn = space.ShiftFn

T = TypeVar('T')
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Simulator = Tuple[InitFn, ApplyFn]



"""Deterministic Simulations

JAX MD includes integrators for deterministic simulations of the NVE, NVT, and
NPT ensembles. For a qualitative description of statistical physics ensembles
see the wikipedia article here:
en.wikipedia.org/wiki/Statistical_ensemble_(mathematical_physics)

Integrators are based direct translation method outlined in the paper,

"A Liouville-operator derived measure-preserving integrator for molecular
dynamics simulations in the isothermal–isobaric ensemble"

M. E. Tuckerman, J. Alejandre, R. López-Rendón, A. L Jochim, and G. J. Martyna
J. Phys. A: Math. Gen. 39 5629 (2006)

As such, we define several primitives that are generically useful in describing
simulations of this type. Namely, the velocity-Verlet integration step that is
used in the NVE and NVT simulations. We also define a general Nose-Hoover chain
primitive that is used to couple components of the system to a chain that
regulates the temperature. These primitives can be combined to construct more
interesting simulations that involve e.g. temperature gradients.
"""


def velocity_verlet(force_fn: Callable[..., Array],
                    shift_fn: ShiftFn,
                    dt: float,
                    state: T,
                    **kwargs) -> T:
  """Apply a single step of velocity Verlet integration to a state."""
  dt = f32(dt)
  dt_2 = f32(dt / 2)
  dt2_2 = f32(dt ** 2 / 2)

  R, V, F, M = state.position, state.velocity, state.force, state.mass

  Minv = 1 / M

  R = shift_fn(R, V * dt + F * dt2_2 * Minv, **kwargs)
  F_new = force_fn(R, **kwargs)
  V += (F + F_new) * dt_2 * Minv
  return dataclasses.replace(state,
                             position=R,
                             velocity=V,
                             force=F_new)


# Constant Energy Simulations


@dataclasses.dataclass
class NVEState:
  """A struct containing the state of an NVE simulation.

  This tuple stores the state of a simulation that samples from the
  microcanonical ensemble in which the (N)umber of particles, the (V)olume, and
  the (E)nergy of the system are held fixed.

  Attributes:
    position: An ndarray of shape `[n, spatial_dimension]` storing the position
      of particles.
    velocity: An ndarray of shape `[n, spatial_dimension]` storing the velocity
      of particles.
    force: An ndarray of shape `[n, spatial_dimension]` storing the force acting
      on particles from the previous step.
    mass: A float or an ndarray of shape `[n]` containing the masses of the
      particles.
  """
  position: Array
  velocity: Array
  force: Array
  mass: Array


# pylint: disable=invalid-name
def nve(energy_or_force_fn: Callable[..., Array],
        shift_fn: ShiftFn,
        dt: float) -> Simulator:
  """Simulates a system in the NVE ensemble.

  Samples from the microcanonical ensemble in which the number of particles
  (N), the system volume (V), and the energy (E) are held constant. We use a
  standard velocity Verlet integration scheme.

  Args:
    energy_or_force: A function that produces either an energy or a force from
      a set of particle positions specified as an ndarray of shape
      `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    quant: Either a quantity. Energy or a quantity. Force specifying whether
      `energy_or_force` is an energy or force respectively.
  Returns:
    See above.
  """
  force_fn = quantity.canonicalize_force(energy_or_force_fn)

  def init_fn(key, R, kT, mass=f32(1.0), **kwargs):
    mass = quantity.canonicalize_mass(mass)
    V = jnp.sqrt(kT / mass) * random.normal(key, R.shape, dtype=R.dtype)
    V = V - jnp.mean(V * mass, axis=0, keepdims=True) / mass
    return NVEState(R, V, force_fn(R, **kwargs), mass)  # pytype: disable=wrong-arg-count

  def step_fn(state, **kwargs):
    return velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)

  return init_fn, step_fn


# Constant Temperature Simulations


# Suzuki-Yoshida weights for integrators of different order.
# These are copied from OpenMM at
# https://github.com/openmm/openmm/blob/master/openmmapi/src/NoseHooverChain.cpp


SUZUKI_YOSHIDA_WEIGHTS = {
    1: [1],
    3: [0.828981543588751, -0.657963087177502, 0.828981543588751],
    5: [0.2967324292201065, 0.2967324292201065, -0.186929716880426,
        0.2967324292201065, 0.2967324292201065],
    7: [0.784513610477560, 0.235573213359357, -1.17767998417887,
        1.31518632068391, -1.17767998417887, 0.235573213359357,
        0.784513610477560]
}


@dataclasses.dataclass
class NoseHooverChain:
  """State information for a Nose-Hoover chain.

  Attributes:
    position: An ndarray of shape `[chain_length]` that stores the position of
      the chain.
    velocity: An ndarray of shape `[chain_length]` that stores the velocity of
      the chain.
    mass: An ndarray of shape `[chain_length]` that stores the mass of the
      chain.
    tau: The desired period of oscillation for the chain. Longer periods result
      is better stability but worse temperature control.
    kinetic_energy: A float that stores the current kinetic energy of the
      system that the chain is coupled to.
    degrees_of_freedom: An integer specifying the number of degrees of freedom
      that the chain is coupled to.
  """
  position: Array
  velocity: Array
  mass: Array
  tau: Array
  kinetic_energy: Array
  degrees_of_freedom: int=dataclasses.static_field()


@dataclasses.dataclass
class NoseHooverChainFns:
  initialize: Callable
  half_step: Callable
  update_mass: Callable


def nose_hoover_chain(dt: float,
                      chain_length: int,
                      chain_steps: int,
                      sy_steps: int,
                      tau: float
                      ) -> NoseHooverChainFns:
  """Helper function to simulate a Nose-Hoover Chain coupled to a system.

  This function is used in simulations that sample from thermal ensembles by
  coupling the system to one, or more, Nose-Hoover chains. We use the direct
  translation method outlined in Martyna et al. [#martyna92]_ and the Nose-Hoover chains are updated
  using two half steps: one at the beginning of a simulation step and one at
  the end. The masses of the Nose-Hoover chains are updated automatically to
  enforce a specific period of oscillation, `tau`. Larger values of `tau` will
  yield systems that reach the target temperature more slowly but are also more
  stable.

  As described in Martyna et al. [#martyna92]_, the Nose-Hoover chain often evolves on a faster
  timescale than the rest of the simulation. Therefore, it sometimes necessary
  to integrate the chain over several substeps for each step of MD. To do this
  we follow the Suzuki-Yoshida scheme. Specifically, we subdivide our chain
  simulation into :math:`n_c` substeps. These substeps are further subdivided into
  :math:`n_sy` steps. Each :math:`n_sy` step has length :math:`\delta_i = \Delta t w_i / n_c`
  where :math:`w_i` are constants such that :math:`\sum_i w_i = 1`. See the table of
  Suzuki-Yoshida weights above for specific values. The number of substeps
  and the number of Suzuki-Yoshida steps are set using the `chain_steps` and
  `sy_steps` arguments.

  Consequently, the Nose-Hoover chains are described by three functions: an
  `init_fn` that initializes the state of the chain, a `half_step_fn` that
  updates the chain for one half-step, and an `update_chain_mass_fn` that
  updates the masses of the chain to enforce the correct period of oscillation.

  Note that a system can have many Nose-Hoover chains coupled to it to produce,
  for example, a temperature gradient. We also note that the NPT ensemble
  naturally features two chains: one that couples to the thermal degrees of
  freedom and one that couples to the barostat.

  Attributes:
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    chain_length: An integer specifying the number of particles in
      the Nose-Hoover chain.
    chain_steps: An integer specifying the number :math:`n_c` of outer substeps.
    sy_steps: An integer specifying the number of Suzuki-Yoshida steps. This
      must be either `1`, `3`, `5`, or `7`.
    tau: A floating point timescale over which temperature equilibration occurs.
      Measured in units of `dt`. The performance of the Nose-Hoover chain
      thermostat can be quite sensitive to this choice.
  Returns:
    A triple of functions that initialize the chain, do a half step of
    simulation, and update the chain masses respectively.
  """

  def init_fn(degrees_of_freedom, KE, kT):
    xi = jnp.zeros(chain_length, KE.dtype)
    v_xi = jnp.zeros(chain_length, KE.dtype)

    Q = kT * tau ** f32(2) * jnp.ones(chain_length, dtype=f32)
    Q = Q.at[0].multiply(degrees_of_freedom)
    return NoseHooverChain(xi, v_xi, Q, tau, KE, degrees_of_freedom)  # pytype: disable=wrong-arg-count

  def substep_fn(delta, V, state, kT):
    """Apply a single update to the chain parameters and rescales velocity."""

    xi, v_xi, Q, _tau, KE, DOF = dataclasses.astuple(state)

    delta_2 = delta   / f32(2.0)
    delta_4 = delta_2 / f32(2.0)
    delta_8 = delta_4 / f32(2.0)

    M = chain_length - 1

    G = (v_xi[M - 1] ** f32(2) * Q[M - 1] - kT) / Q[M]
    v_xi = v_xi.at[M].add(delta_4 * G)

    def backward_loop_fn(v_xi_new, m):
      G = (v_xi[m - 1] ** 2 * Q[m - 1] - kT) / Q[m]
      scale = jnp.exp(-delta_8 * v_xi_new)
      v_xi_new = scale * (scale * v_xi[m] + delta_4 * G)
      return v_xi_new, v_xi_new
    idx = jnp.arange(M - 1, 0, -1)
    _, v_xi_update = lax.scan(backward_loop_fn, v_xi[M], idx, unroll=2)
    v_xi = v_xi.at[idx].set(v_xi_update)

    G = (f32(2.0) * KE - DOF * kT) / Q[0]
    scale = jnp.exp(-delta_8 * v_xi[1])
    v_xi = v_xi.at[0].set(scale * (scale * v_xi[0] + delta_4 * G))

    scale = jnp.exp(-delta_2 * v_xi[0])
    KE = KE * scale ** f32(2)
    V = V * scale

    xi = xi + delta_2 * v_xi

    G = (f32(2) * KE - DOF * kT) / Q[0]
    def forward_loop_fn(G, m):
      scale = jnp.exp(-delta_8 * v_xi[m + 1])
      v_xi_update = scale * (scale * v_xi[m] + delta_4 * G)
      G = (v_xi_update ** 2 * Q[m] - kT) / Q[m + 1]
      return G, v_xi_update
    idx = jnp.arange(M)
    G, v_xi_update = lax.scan(forward_loop_fn, G, idx, unroll=2)
    v_xi = v_xi.at[idx].set(v_xi_update)
    v_xi = v_xi.at[M].add(delta_4 * G)

    return V, NoseHooverChain(xi, v_xi, Q, _tau, KE, DOF), kT  # pytype: disable=wrong-arg-count

  def half_step_chain_fn(V, state, kT):
    if chain_steps == 1 and sy_steps == 1:
      V, state, _ = substep_fn(dt, V, state, kT)
      return V, state

    delta = dt / chain_steps
    ws = jnp.array(SUZUKI_YOSHIDA_WEIGHTS[sy_steps], dtype=V.dtype)
    def body_fn(cs, i):
      d = f32(delta * ws[i % sy_steps])
      return substep_fn(d, *cs), 0
    V, state, _ = lax.scan(body_fn,
                           (V, state, kT),
                           jnp.arange(chain_steps * sy_steps))[0]
    return V, state

  def update_chain_mass_fn(state, kT):
    xi, v_xi, Q, _tau, KE, DOF = dataclasses.astuple(state)

    Q = kT * _tau ** f32(2) * jnp.ones(chain_length, dtype=f32)
    Q = Q.at[0].multiply(DOF)

    return NoseHooverChain(xi, v_xi, Q, _tau, KE, DOF)  # pytype: disable=wrong-arg-count

  return NoseHooverChainFns(init_fn, half_step_chain_fn, update_chain_mass_fn)  # pytype: disable=wrong-arg-count


def default_nhc_kwargs(tau: float, overrides: Dict) -> Dict:
  default_kwargs = {
      'chain_length': 3,
      'chain_steps': 2,
      'sy_steps': 3,
      'tau': tau
  }

  if overrides is None:
    return default_kwargs

  return {
      key: overrides.get(key, default_kwargs[key])
      for key in default_kwargs
  }


@dataclasses.dataclass
class NVTNoseHooverState:
  """State information for an NVT system with a Nose-Hoover chain thermostat.

  Attributes:
    position: The current position of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    velocity: The velocity of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    force: The current force on the particles. An ndarray of floats with shape
      `[n, spatial_dimension]`.
    mass: The mass of the particles. Can either be a float or an ndarray
      of floats with shape `[n]`.
    chain: The variables describing the Nose-Hoover chain.
  """
  position: Array
  velocity: Array
  force: Array
  mass: Array
  chain: NoseHooverChain


def nvt_nose_hoover(energy_or_force_fn: Callable[..., Array],
                    shift_fn: ShiftFn,
                    dt: float,
                    kT: float,
                    chain_length: int=5,
                    chain_steps: int=2,
                    sy_steps: int=3,
                    tau: Optional[float]=None) -> Simulator:
  """Simulation in the NVT ensemble using a Nose Hoover Chain thermostat.

  Samples from the canonical ensemble in which the number of particles (N),
  the system volume (V), and the temperature (T) are held constant. We use a
  Nose Hoover Chain (NHC) thermostat described in [#martyna92]_ [#martyna98]_ [#tuckerman]_ . 
  We follow the direct translation method outlined in Tuckerman et al. [#tuckerman]_ 
  and the interested reader might want to look at that paper as a reference.

  Args:
    energy_or_force: A function that produces either an energy or a force from
      a set of particle positions specified as an ndarray of shape
      `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    kT: Floating point number specifying the temperature in units of Boltzmann
      constant. To update the temperature dynamically during a simulation one
      should pass `kT` as a keyword argument to the step function.
    chain_length: An integer specifying the number of particles in
      the Nose-Hoover chain.
    chain_steps: An integer specifying the number, :math:`n_c`, of outer substeps.
    sy_steps: An integer specifying the number of Suzuki-Yoshida steps. This
      must be either `1`, `3`, `5`, or `7`.
    tau: A floating point timescale over which temperature equilibration occurs.
      Measured in units of `dt`. The performance of the Nose-Hoover chain
      thermostat can be quite sensitive to this choice.
  Returns:
    See above.

  .. rubric:: References
  .. [#martyna92] Martyna, Glenn J., Michael L. Klein, and Mark Tuckerman.
    "Nose-Hoover chains: The canonical ensemble via continuous dynamics."
    The Journal of chemical physics 97, no. 4 (1992): 2635-2643.
  .. [#martyna98] Martyna, Glenn, Mark Tuckerman, Douglas J. Tobias, and Michael L. Klein.
    "Explicit reversible integrators for extended systems dynamics."
    Molecular Physics 87. (1998) 1117-1157.
  .. [#tuckerman] Tuckerman, Mark E., Jose Alejandre, Roberto Lopez-Rendon,
    Andrea L. Jochim, and Glenn J. Martyna.
    "A Liouville-operator derived measure-preserving integrator for molecular
    dynamics simulations in the isothermal-isobaric ensemble."
    Journal of Physics A: Mathematical and General 39, no. 19 (2006): 5629.
  """
  dt = f32(dt)
  if tau is None:
    tau = dt * 100
  tau = f32(tau)

  force_fn = quantity.canonicalize_force(energy_or_force_fn)
  chain_fns = nose_hoover_chain(dt, chain_length, chain_steps, sy_steps, tau)

  def init_fn(key, R, mass=f32(1.0), **kwargs):
    _kT = kT if 'kT' not in kwargs else kwargs['kT']

    mass = quantity.canonicalize_mass(mass)
    V = jnp.sqrt(_kT / mass) * random.normal(key, R.shape, dtype=R.dtype)
    V = V - jnp.mean(V * mass, axis=0, keepdims=True) / mass
    KE = quantity.kinetic_energy(V, mass)

    return NVTNoseHooverState(R, V, force_fn(R, **kwargs), mass,
                              chain_fns.initialize(R.size, KE, _kT))  # pytype: disable=wrong-arg-count
  def apply_fn(state, **kwargs):
    _kT = kT if 'kT' not in kwargs else kwargs['kT']

    chain = state.chain

    chain = chain_fns.update_mass(chain, _kT)

    v, chain = chain_fns.half_step(state.velocity, chain, _kT)
    state = dataclasses.replace(state, velocity=v)

    state = velocity_verlet(force_fn, shift_fn, dt, state, **kwargs)

    KE = quantity.kinetic_energy(state.velocity, state.mass)
    chain = dataclasses.replace(chain, kinetic_energy=KE)

    v, chain = chain_fns.half_step(state.velocity, chain, _kT)
    state = dataclasses.replace(state, velocity=v, chain=chain)

    return state
  return init_fn, apply_fn


def nvt_nose_hoover_invariant(energy_fn: Callable[..., Array],
                              state: NVTNoseHooverState,
                              kT: float,
                              **kwargs) -> float:
  """The conserved quantity for the NVT ensemble with a Nose-Hoover thermostat.

  This function is normally used for debugging the Nose-Hoover thermostat.

  Arguments:
    energy_fn: The energy function of the Nose-Hoover system.
    state: The current state of the system.
    kT: The current goal temperature of the system.

  Returns:
    The Hamiltonian of the extended NVT dynamics.
  """
  PE = energy_fn(state.position, **kwargs)
  KE = quantity.kinetic_energy(state.velocity, state.mass)

  DOF = state.position.size
  E = PE + KE

  c = state.chain

  E += c.mass[0] * c.velocity[0] ** 2 / 2 + DOF * kT * c.position[0]
  for r, v, m in zip(c.position[1:], c.velocity[1:], c.mass[1:]):
    E += m * v ** 2 / 2 + kT * r
  return E


@dataclasses.dataclass
class NPTNoseHooverState:
  """State information for an NPT system with Nose-Hoover chain thermostats.

  Attributes:
    position: The current position of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    velocity: The velocity of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    force: The current force on the particles. An ndarray of floats with shape
      `[n, spatial_dimension]`.
    mass: The mass of the particles. Can either be a float or an ndarray
      of floats with shape `[n]`.
    reference_box: A box used to measure relative changes to the simulation
      environment.
    box_position: A positional degree of freedom used to describe the current
      box. The box_position is parameterized as `box_position = (1/d)log(V/V_0)`
      where `V` is the current volume, `V_0` is the reference volume and `d` is
      the spatial dimension.
    box_velocity: A velocity degree of freedom for the box.
    box_mass: The mass assigned to the box.
    barostat: The variables describing the Nose-Hoover chain coupled to the
      barostat.
    thermostsat: The variables describing the Nose-Hoover chain coupled to the
      thermostat.
  """
  position: Array
  velocity: Array
  force: Array
  mass: Array

  reference_box: Box

  box_position: Array
  box_velocity: Array
  box_mass: Array

  barostat: NoseHooverChain
  thermostat: NoseHooverChain


def _npt_box_info(state: NPTNoseHooverState) -> Tuple[float,
                                                      Callable[[float], float]]:
  """Gets the current volume and a function to compute the box from volume."""
  dim = state.position.shape[1]
  ref = state.reference_box
  V_0 = quantity.volume(dim, ref)
  V = V_0 * jnp.exp(dim * state.box_position)
  return V, lambda V: (V / V_0) ** (1 / dim) * ref


def npt_box(state: NPTNoseHooverState) -> Box:
  """Get the current box from an NPT simulation."""
  dim = state.position.shape[1]
  ref = state.reference_box
  V_0 = quantity.volume(dim, ref)
  V = V_0 * jnp.exp(dim * state.box_position)
  return (V / V_0) ** (1 / dim) * ref


def npt_nose_hoover(energy_fn: Callable[..., Array],
                    shift_fn: ShiftFn,
                    dt: float,
                    pressure: float,
                    kT: float,
                    barostat_kwargs: Optional[Dict]=None,
                    thermostat_kwargs: Optional[Dict]=None) -> Simulator:
  """Simulation in the NPT ensemble using a pair of Nose Hoover Chains.

  Samples from the canonical ensemble in which the number of particles (N),
  the system pressure (P), and the temperature (T) are held constant. 
  We use a pair of Nose Hoover Chains (NHC) described in 
  [#martyna92]_ [#martyna98]_ [#tuckerman]_ coupled to the
  barostat and the thermostat respectively. We follow the direct translation
  method outlined in Tuckerman et al. [#tuckerman]_ and the interested reader 
  might want to look at that paper as a reference.

  Args:
    energy_fn: A function that produces either an energy from a set of particle
      positions specified as an ndarray of shape `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    pressure: Floating point number specifying the target pressure. To update
      the pressure dynamically during a simulation one should pass `pressure`
      as a keyword argument to the step function.
    kT: Floating point number specifying the temperature in units of Boltzmann
      constant. To update the temperature dynamically during a simulation one
      should pass `kT` as a keyword argument to the step function.
    barostat_kwargs: A dictionary of keyword arguments passed to the barostat
      NHC. Any parameters not set are drawn from a relatively robust default
      set.
    thermostat_kwargs: A dictionary of keyword arguments passed to the
      thermostat NHC. Any parameters not set are drawn from a relatively robust
      default set.

  Returns:
    See above.

  """

  t = f32(dt)
  dt_2 = f32(dt / 2)

  force_fn = quantity.force(energy_fn)

  barostat_kwargs = default_nhc_kwargs(1000 * dt, barostat_kwargs)
  barostat = nose_hoover_chain(dt, **barostat_kwargs)

  thermostat_kwargs = default_nhc_kwargs(100 * dt, thermostat_kwargs)
  thermostat = nose_hoover_chain(dt, **thermostat_kwargs)

  def init_fn(key, R, box, mass=f32(1.0), **kwargs):
    N, dim = R.shape

    _kT = kT if 'kT' not in kwargs else kwargs['kT']

    mass = quantity.canonicalize_mass(mass)
    V = jnp.sqrt(_kT / mass) * random.normal(key, R.shape, dtype=R.dtype)
    V = V - jnp.mean(V * mass, axis=0, keepdims=True) / mass
    KE = quantity.kinetic_energy(V, mass)

    # The box position is defined via pos = (1 / d) log V / V_0.
    zero = jnp.zeros((), dtype=R.dtype)
    one = jnp.ones((), dtype=R.dtype)
    box_position = zero
    box_velocity = zero
    box_mass = dim * (N + 1) * kT * barostat_kwargs['tau'] ** 2 * one
    KE_box = quantity.kinetic_energy(box_velocity, box_mass)

    if jnp.isscalar(box) or box.ndim == 0:
      # TODO(schsam): This is necessary because of JAX issue #5849.
      box = jnp.eye(R.shape[-1]) * box

    return NPTNoseHooverState(R, V, force_fn(R, box=box, **kwargs), mass, box,
                              box_position, box_velocity, box_mass,
                              barostat.initialize(1, KE_box, _kT),
                              thermostat.initialize(R.size, KE, _kT))  # pytype: disable=wrong-arg-count

  def update_box_mass(state, kT):
    N, dim = state.position.shape
    dtype = state.position.dtype
    box_mass = jnp.array(dim * (N + 1) * kT * state.barostat.tau ** 2, dtype)
    return dataclasses.replace(state, box_mass=box_mass)

  def box_force(alpha, vol, box_fn, position, velocity, mass, force, pressure,
                **kwargs):
    N, dim = position.shape

    def U(eps):
      return energy_fn(position, box=box_fn(vol), perturbation=(1+eps), **kwargs)

    dUdV = grad(U)
    KE2 = util.high_precision_sum(velocity ** 2 * mass)
    R = space.transform(box_fn(vol), position)
    RdotF = util.high_precision_sum(R * force)

    return alpha * KE2 + RdotF - dUdV(0.0) - pressure * vol * dim

  def sinhx_x(x):
    """Taylor series for sinh(x) / x as x -> 0."""
    return (1 + x ** 2 / 6 + x ** 4 / 120 + x ** 6 / 5040 +
            x ** 8 / 362_880 + x ** 10 / 39_916_800)

  def exp_iL1(box, R, V, V_b, **kwargs):
    x = V_b * dt
    x_2 = x / 2
    sinhV = sinhx_x(x_2)  # jnp.sinh(x_2) / x_2
    return shift_fn(R, R * (jnp.exp(x) - 1) + dt * V * jnp.exp(x_2) * sinhV,
                    box=box, **kwargs)  # pytype: disable=wrong-keyword-args

  def exp_iL2(alpha, V, A, V_b):
    x = alpha * V_b * dt_2
    x_2 = x / 2
    sinhV = sinhx_x(x_2)  # jnp.sinh(x_2) / x_2
    return V * jnp.exp(-x) + dt_2 * A * sinhV * jnp.exp(-x_2)

  def inner_step(state, **kwargs):
    _pressure = kwargs.pop('pressure', pressure)

    R, V, M, F = state.position, state.velocity, state.mass, state.force
    R_b, V_b, M_b = state.box_position, state.box_velocity, state.box_mass

    N, dim = R.shape

    vol, box_fn = _npt_box_info(state)

    alpha = 1 + 1 / N
    G_e = box_force(alpha, vol, box_fn, R, V, M, F, _pressure, **kwargs)
    V_b = V_b + dt_2 * G_e / M_b
    V = exp_iL2(alpha, V, F / M, V_b)

    R_b = R_b + V_b * dt
    state = dataclasses.replace(state, box_position=R_b)

    vol, box_fn = _npt_box_info(state)

    box = box_fn(vol)
    R = exp_iL1(box, R, V, V_b)
    F = force_fn(R, box=box, **kwargs)

    V = exp_iL2(alpha, V, F / M, V_b)
    G_e = box_force(alpha, vol, box_fn, R, V, M, F, _pressure, **kwargs)
    V_b = V_b + dt_2 * G_e / M_b

    return dataclasses.replace(state,
                               position=R, velocity=V, mass=M, force=F,
                               box_position=R_b, box_velocity=V_b, box_mass=M_b)

  def apply_fn(state, **kwargs):
    S = state
    _kT = kT if 'kT' not in kwargs else kwargs['kT']

    bc = barostat.update_mass(S.barostat, _kT)
    tc = thermostat.update_mass(S.thermostat, _kT)
    S = update_box_mass(S, _kT)

    V_b, bc = barostat.half_step(S.box_velocity, bc, _kT)
    V, tc = thermostat.half_step(S.velocity, tc, _kT)

    S = dataclasses.replace(S, velocity=V, box_velocity=V_b)
    S = inner_step(S, **kwargs)

    KE = quantity.kinetic_energy(S.velocity, S.mass)
    tc = dataclasses.replace(tc, kinetic_energy=KE)

    KE_box = quantity.kinetic_energy(S.box_velocity, S.box_mass)
    bc = dataclasses.replace(bc, kinetic_energy=KE_box)

    V, tc = thermostat.half_step(S.velocity, tc, _kT)
    V_b, bc = barostat.half_step(S.box_velocity, bc, _kT)

    S = dataclasses.replace(S,
                            thermostat=tc, barostat=bc,
                            velocity=V, box_velocity=V_b)

    return S
  return init_fn, apply_fn


def npt_nose_hoover_invariant(energy_fn: Callable[..., Array],
                              state: NPTNoseHooverState,
                              pressure: float,
                              kT: float,
                              **kwargs) -> float:
  """The conserved quantity for the NPT ensemble with a Nose-Hoover thermostat.

  This function is normally used for debugging the NPT simulation.

  Arguments:
    energy_fn: The energy function of the system.
    state: The current state of the system.
    pressure: The current goal pressure of the system.
    kT: The current goal temperature of the system.

  Returns:
    The Hamiltonian of the extended NPT dynamics.
  """
  volume, box_fn = _npt_box_info(state)
  PE = energy_fn(state.position, box=box_fn(volume), **kwargs)
  KE = quantity.kinetic_energy(state.velocity, state.mass)

  DOF = state.position.size
  E = PE + KE

  c = state.thermostat
  E += c.mass[0] * c.velocity[0] ** 2 / 2 + DOF * kT * c.position[0]
  for r, v, m in zip(c.position[1:], c.velocity[1:], c.mass[1:]):
    E += m * v ** 2 / 2 + kT * r

  c = state.barostat
  for r, v, m in zip(c.position, c.velocity, c.mass):
    E += m * v ** 2 / 2 + kT * r

  E += pressure * volume
  E += state.box_mass * state.box_velocity ** 2 / 2

  return E


"""Stochastic Simulations

JAX MD includes integrators for stochastic simulations of Langevin dynamics and
Brownian motion for systems in the NVT ensemble with a solvent.
"""

@dataclasses.dataclass
class NVTLangevinState:
  """A struct containing state information for the Langevin thermostat.

  Attributes:
    position: The current position of the particles. An ndarray of floats with
      shape `[n, spatial_dimension]`.
    velocity: The velocity of particles. An ndarray of floats with shape
      `[n, spatial_dimension]`.
    force: The (non-stochastic) force on particles. An ndarray of floats with
      shape `[n, spatial_dimension]`.
    mass: The mass of particles. Will either be a float or an ndarray of floats
      with shape `[n]`.
    rng: The current state of the random number generator.
  """
  position: Array
  velocity: Array
  force: Array
  mass: Array
  rng: Array

def nvt_langevin(energy_or_force: Callable[..., Array],
                 shift: ShiftFn,
                 dt: float,
                 kT: float,
                 gamma: float=0.1,
                 center_velocity: bool=True) -> Simulator:
  """Simulation in the NVT ensemble using the Langevin thermostat.

  Samples from the canonical ensemble in which the number of particles (N),
  the system volume (V), and the temperature (T) are held constant. Langevin
  dynamics are stochastic and it is supposed that the system is interacting with
  fictitious microscopic degrees of freedom. An example of this would be large
  particles in a solvent such as water. Thus, Langevin dynamics are a stochastic
  ODE described by a friction coefficient and noise of a given covariance.

  Our implementation follows the excellent set of lecture notes by Carlon,
  Laleman, and Nomidis [#carlon]_ .

  Args:
    energy_or_force: A function that produces either an energy or a force from
      a set of particle positions specified as an ndarray of shape
      `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    kT: Floating point number specifying the temperature in units of Boltzmann
      constant. To update the temperature dynamically during a simulation one
      should pass `kT` as a keyword argument to the step function.
    gamma: A float specifying the friction coefficient between the particles
      and the solvent.
    center_velocity: A boolean specifying whether or not the center of mass
      position should be subtracted.
  Returns:
    See above.

  .. rubric:: References
  .. [#carlon] E. Carlon, M. Laleman, S. Nomidis. "Molecular Dynamics Simulation."
    http://itf.fys.kuleuven.be/~enrico/Teaching/molecular_dynamics_2015.pdf
    Accessed on 06/05/2019.
  """

  force_fn = quantity.canonicalize_force(energy_or_force)

  dt_2 = f32(dt / 2)
  dt2 = f32(dt ** 2 / 2)
  dt32 = f32(dt ** (3.0 / 2.0) / 2)

  kT = f32(kT)

  gamma = f32(gamma)

  def init_fn(key, R, mass=f32(1), **kwargs):
    _kT = kT if 'kT' not in kwargs else kwargs['kT']
    mass = quantity.canonicalize_mass(mass)

    key, split = random.split(key)

    V = jnp.sqrt(_kT / mass) * random.normal(split, R.shape, dtype=R.dtype)
    if center_velocity:
      V = V - jnp.mean(V, axis=0, keepdims=True)

    return NVTLangevinState(R, V, force_fn(R, **kwargs) / mass, mass, key)  # pytype: disable=wrong-arg-count

  def apply_fn(state, **kwargs):
    R, V, F, mass, key = dataclasses.astuple(state)

    _kT = kT if 'kT' not in kwargs else kwargs['kT']
    N, dim = R.shape

    key, xi_key, theta_key = random.split(key, 3)
    xi = random.normal(xi_key, (N, dim), dtype=R.dtype)
    sqrt3 = f32(jnp.sqrt(3))
    theta = random.normal(theta_key, (N, dim), dtype=R.dtype) / sqrt3

    # NOTE(schsam): We really only need to recompute sigma if the temperature
    # is nonconstant. @Optimization
    # TODO(schsam): Check that this is really valid in the case that the masses
    # are non identical for all particles.
    sigma = jnp.sqrt(f32(2) * _kT * gamma / mass)
    C = dt2 * (F - gamma * V) + sigma * dt32 * (xi + theta)

    R = shift(R, dt * V + C, **kwargs)
    F_new = force_fn(R, **kwargs) / mass
    V = (f32(1) - dt * gamma) * V + dt_2 * (F_new + F)
    V = V + sigma * jnp.sqrt(dt) * xi - gamma * C

    return NVTLangevinState(R, V, F_new, mass, key)  # pytype: disable=wrong-arg-count
  return init_fn, apply_fn


@dataclasses.dataclass
class BrownianState:
  """A tuple containing state information for Brownian dynamics.

  Attributes:
    position: The current position of the particles. An ndarray of floats with
      shape `[n, spatial_dimension]`.
    mass: The mass of particles. Will either be a float or an ndarray of floats
      with shape `[n]`.
    rng: The current state of the random number generator.
  """
  position: Array
  mass: Array
  rng: Array


def brownian(energy_or_force: Callable[..., Array],
             shift: ShiftFn,
             dt: float,
             kT: float,
             gamma: float=0.1) -> Simulator:
  """Simulation of Brownian dynamics.

  Simulates Brownian dynamics which are synonymous with the overdamped
  regime of Langevin dynamics. However, in this case we don't need to take into
  account velocity information and the dynamics simplify. Consequently, when
  Brownian dynamics can be used they will be faster than Langevin. As in the
  case of Langevin dynamics our implementation follows Carlon et al. [#carlon]_

  Args:
    energy_or_force: A function that produces either an energy or a force from
      a set of particle positions specified as an ndarray of shape
      `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    dt: Floating point number specifying the timescale (step size) of the
      simulation.
    kT: Floating point number specifying the temperature in units of Boltzmann
      constant. To update the temperature dynamically during a simulation one
      should pass `kT` as a keyword argument to the step function.
    gamma: A float specifying the friction coefficient between the particles
      and the solvent.

  Returns:
    See above.
  """

  force_fn = quantity.canonicalize_force(energy_or_force)

  dt, gamma = static_cast(dt, gamma)

  def init_fn(key, R, mass=f32(1)):
    mass = quantity.canonicalize_mass(mass)

    return BrownianState(R, mass, key)  # pytype: disable=wrong-arg-count

  def apply_fn(state, **kwargs):
    _kT = kT if 'kT' not in kwargs else kwargs['kT']

    R, mass, key = dataclasses.astuple(state)

    key, split = random.split(key)

    F = force_fn(R, **kwargs)
    xi = random.normal(split, R.shape, R.dtype)

    nu = f32(1) / (mass * gamma)

    dR = F * dt * nu + jnp.sqrt(f32(2) * _kT * dt * nu) * xi
    R = shift(R, dR, **kwargs)

    return BrownianState(R, mass, key)  # pytype: disable=wrong-arg-count

  return init_fn, apply_fn


"""Experimental Simulations.


Below are simulation environments whose implementation is somewhat
experimental / preliminary. These environments might not be as ergonomic
as the more polished environments above.
"""


@dataclasses.dataclass
class SwapMCState:
  """A struct containing state information about a Hybrid Swap MC simulation.

  Attributes:
    md: A NVTNoseHooverState containing continuous molecular dynamics data.
    sigma: An `[n,]` array of particle radii.
    key: A JAX PRGNKey used for random number generation.
    neighbor: A NeighborList for the system.
  """
  md: NVTNoseHooverState
  sigma: Array
  key: Array
  neighbor: partition.NeighborList


# pytype: disable=wrong-arg-count
# pytype: disable=wrong-keyword-args
def hybrid_swap_mc(space_fns: space.Space,
                   energy_fn: Callable[[Array, Array], Array],
                   neighbor_fn: partition.NeighborFn,
                   dt: float,
                   kT: float,
                   t_md: float,
                   N_swap: int,
                   sigma_fn: Optional[Callable[[Array], Array]]=None
                   ) -> Simulator:
  """Simulation of Hybrid Swap Monte-Carlo.

  This code simulates the hybrid Swap Monte Carlo algorithm introduced in 
  Berthier et al. [#berthier]_
  Here an NVT simulation is performed for `t_md` time and then `N_swap` MC
  moves are performed that swap the radii of randomly chosen particles. The
  random swaps are accepted with Metropolis-Hastings step. Each call to the
  step function runs molecular dynamics for `t_md` and then performs the swaps.

  Note that this code doesn't feature some of the convenience functions in the
  other simulations. In particular, there is no support for dynamics keyword
  arguments and the energy function must be a simple callable of two variables:
  the distance between adjacent particles and the diameter of the particles.
  If you want support for a better notion of potential or dynamic keyword
  arguments, please file an issue!

  Args:
    space_fns: A tuple of a displacement function and a shift function defined
      in `space.py`.
    energy_fn: A function that computes the energy between one pair of
      particles as a function of the distance between the particles and the
      diameter. This function should not have been passed to `smap.xxx`.
    neighbor_fn: A function to construct neighbor lists outlined in
      `partition.py`.
    dt: The timestep used for the continuous time MD portion of the simulation.
    kT: The temperature of heat bath that the system is coupled to during MD.
    t_md: The time of each MD block.
    N_swap: The number of swapping moves between MD blocks.
    sigma_fn: An optional function for combining radii if they are to be
      non-additive.

  Returns:
    See above.

  .. rubric:: References
  .. [#berthier] L. Berthier, E. Flenner, C. J. Fullerton, C. Scalliet, and M. Singh.
    "Efficient swap algorithms for molecular dynamics simulations of
    equilibrium supercooled liquids", J. Stat. Mech. (2019) 064004
  """
  displacement_fn, shift_fn = space_fns
  metric_fn = space.metric(displacement_fn)
  nbr_metric_fn = space.map_neighbor(metric_fn)

  md_steps = int(t_md // dt)

  # Canonicalize the argument names to be dr and sigma.
  wrapped_energy_fn = lambda dr, sigma: energy_fn(dr, sigma)
  if sigma_fn is None:
    sigma_fn = lambda si, sj: 0.5 * (si + sj)
  nbr_energy_fn = smap.pair_neighbor_list(wrapped_energy_fn,
                                          metric_fn,
                                          sigma=sigma_fn)

  nvt_init_fn, nvt_step_fn = nvt_nose_hoover(nbr_energy_fn,
                                             shift_fn,
                                             dt,
                                             kT=kT,
                                             chain_length=3)
  def init_fn(key, position, sigma, nbrs=None):
    key, sim_key = random.split(key)
    nbrs = neighbor_fn(position, nbrs)  # pytype: disable=wrong-arg-count
    md_state = nvt_init_fn(sim_key, position, neighbor=nbrs, sigma=sigma)
    return SwapMCState(md_state, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

  def md_step_fn(i, state):
    md, sigma, key, nbrs = dataclasses.unpack(state)
    md = nvt_step_fn(md, neighbor=nbrs, sigma=sigma)  # pytype: disable=wrong-keyword-args
    nbrs = neighbor_fn(md.position, nbrs)
    return SwapMCState(md, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

  def swap_step_fn(i, state):
    md, sigma, key, nbrs = dataclasses.unpack(state)

    N = md.position.shape[0]

    # Swap a random pair of particle radii.
    key, particle_key, accept_key = random.split(key, 3)
    ij = random.randint(particle_key, (2,), jnp.array(0), jnp.array(N))
    new_sigma = sigma.at[ij].set([sigma[ij[1]], sigma[ij[0]]])

    # Collect neighborhoods around the two swapped particles.
    nbrs_ij = nbrs.idx[ij]
    R_ij = md.position[ij]
    R_neigh = md.position[nbrs_ij]

    sigma_ij = sigma[ij][:, None]
    sigma_neigh = sigma[nbrs_ij]

    new_sigma_ij = new_sigma[ij][:, None]
    new_sigma_neigh = new_sigma[nbrs_ij]

    dR = nbr_metric_fn(R_ij, R_neigh)

    # Compute the energy before the swap.
    energy = energy_fn(dR, sigma_fn(sigma_ij, sigma_neigh))
    energy = jnp.sum(energy * (nbrs_ij < N))

    # Compute the energy after the swap.
    new_energy = energy_fn(dR, sigma_fn(new_sigma_ij, new_sigma_neigh))
    new_energy = jnp.sum(new_energy * (nbrs_ij < N))

    # Accept or reject with a metropolis probability.
    p = random.uniform(accept_key, ())
    accept_prob = jnp.minimum(1, jnp.exp(-(new_energy - energy) / kT))
    sigma = jnp.where(p < accept_prob, new_sigma, sigma)

    return SwapMCState(md, sigma, key, nbrs)  # pytype: disable=wrong-arg-count

  def block_fn(state):
    state = lax.fori_loop(0, md_steps, md_step_fn, state)
    state = lax.fori_loop(0, N_swap, swap_step_fn, state)
    return state

  return init_fn, block_fn
# pytype: enable=wrong-arg-count
# pytype: enable=wrong-keyword-args



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines testing utility functions."""

from absl import flags
from absl.testing import parameterized

import os

from typing import Dict, Sequence

import netCDF4 as nc

import jax.numpy as jnp
import numpy as onp
from jax.config import config

import jax
from jax import dtypes as _dtypes
from jax import jit
from jax import vmap

from jax_md import dataclasses

flags.DEFINE_string(
    'jax_test_dut',
    '',
    help=
    'Describes the device under test in case special consideration is required.'
)


FLAGS = flags.FLAGS


f32 = jnp.float32


# Utility functions forked from jax._src.public_test_util


_python_scalar_dtypes : dict = {
    bool: onp.dtype('bool'),
    int: onp.dtype('int64'),
    float: onp.dtype('float64'),
    complex: onp.dtype('complex128'),
}


def _dtype(x):
  if hasattr(x, 'dtype'):
    return x.dtype
  elif type(x) in _python_scalar_dtypes:
    return onp.dtype(_python_scalar_dtypes[type(x)])
  else:
    return onp.asarray(x).dtype


def is_sequence(x):
  try:
    iter(x)
  except TypeError:
    return False
  else:
    return True


def device_under_test():
  return getattr(FLAGS, 'jax_test_dut', None) or jax.default_backend()


_DEFAULT_TOLERANCE = {
    onp.dtype(onp.bool_): 0,
    onp.dtype(onp.int32): 0,
    onp.dtype(onp.int64): 0,
    onp.dtype(onp.float32): 1e-6,
    onp.dtype(onp.float64): 1e-15,
}


def _default_tolerance():
  if device_under_test() != 'tpu':
    return _DEFAULT_TOLERANCE
  tol = _DEFAULT_TOLERANCE.copy()
  tol[onp.dtype(onp.float32)] = 5e-2
  tol[onp.dtype(onp.complex64)] = 5e-2
  return tol


def _assert_numpy_allclose(a, b, atol=None, rtol=None, err_msg=''):
  if a.dtype == b.dtype == _dtypes.float0:
    onp.testing.assert_array_equal(a, b, err_msg=err_msg)
    return
  a = a.astype(onp.float32) if a.dtype == _dtypes.bfloat16 else a
  b = b.astype(onp.float32) if b.dtype == _dtypes.bfloat16 else b
  kw = {}
  if atol: kw['atol'] = atol
  if rtol: kw['rtol'] = rtol
  with onp.errstate(invalid='ignore'):
    # TODO(phawkins): surprisingly, assert_allclose sometimes reports invalid
    # value errors. It should not do that.
    onp.testing.assert_allclose(a, b, **kw, err_msg=err_msg)


def _tolerance(dtype, tol=None):
  tol = {} if tol is None else tol
  if not isinstance(tol, dict):
    return tol
  tol = {onp.dtype(key): value for key, value in tol.items()}
  dtype = _dtypes.canonicalize_dtype(onp.dtype(dtype))
  return tol.get(dtype, _default_tolerance()[dtype])


_CACHED_INDICES: Dict[int, Sequence[int]] = {}


def cases_from_list(xs):
  xs = list(xs)
  return xs


class JAXMDTestCase(parameterized.TestCase):
  """Testing helper class forked from JaxTestCase."""

  def _assertAllClose(self, x, y, *, check_dtypes=True, atol=None, rtol=None,
                      canonicalize_dtypes=True, err_msg=''):
    """Assert that x and y, either arrays or nested tuples/lists, are close."""
    if isinstance(x, dict):
      self.assertIsInstance(y, dict)
      self.assertEqual(set(x.keys()), set(y.keys()))
      for k in x.keys():
        self._assertAllClose(x[k], y[k], check_dtypes=check_dtypes, atol=atol,
                             rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
                             err_msg=err_msg)
    elif is_sequence(x) and not hasattr(x, '__array__'):
      self.assertTrue(is_sequence(y) and not hasattr(y, '__array__'))
      self.assertEqual(len(x), len(y))
      for x_elt, y_elt in zip(x, y):
        self._assertAllClose(x_elt, y_elt, check_dtypes=check_dtypes, atol=atol,
                             rtol=rtol, canonicalize_dtypes=canonicalize_dtypes,
                             err_msg=err_msg)
    elif hasattr(x, '__array__') or onp.isscalar(x):
      self.assertTrue(hasattr(y, '__array__') or onp.isscalar(y))
      if check_dtypes:
        self.assertDtypesMatch(x, y, canonicalize_dtypes=canonicalize_dtypes)
      x = onp.asarray(x)
      y = onp.asarray(y)
      self.assertArraysAllClose(x, y, check_dtypes=False, atol=atol, rtol=rtol,
                                err_msg=err_msg)
    elif x == y:
      return
    else:
      raise TypeError((type(x), type(y)))

  def assertArraysAllClose(self, x, y, *, check_dtypes=True, atol=None,
                           rtol=None, err_msg=''):
    """Assert that x and y are close (up to numerical tolerances)."""
    self.assertEqual(x.shape, y.shape)
    atol = max(_tolerance(_dtype(x), atol), _tolerance(_dtype(y), atol))
    rtol = max(_tolerance(_dtype(x), rtol), _tolerance(_dtype(y), rtol))
    _assert_numpy_allclose(x, y, atol=atol, rtol=rtol, err_msg=err_msg)

    if check_dtypes:
      self.assertDtypesMatch(x, y)

  def assertDtypesMatch(self, x, y, *, canonicalize_dtypes=True):
    if not config.x64_enabled and canonicalize_dtypes:
      self.assertEqual(_dtypes.canonicalize_dtype(_dtype(x)),
                       _dtypes.canonicalize_dtype(_dtype(y)))
    else:
      self.assertEqual(_dtype(x), _dtype(y))

  def assertAllClose(
      self,
      x,
      y,
      *,
      check_dtypes=True,
      atol=None,
      rtol=None,
      canonicalize_dtypes=True,
      err_msg=''):
    def is_finite(x):
      self.assertTrue(jnp.all(jnp.isfinite(x)))

    jax.tree_map(is_finite, x)
    jax.tree_map(is_finite, y)

    def assert_close(x, y):
      self._assertAllClose(
          x, y,
          check_dtypes=check_dtypes,
          atol=atol,
          rtol=rtol,
          canonicalize_dtypes=canonicalize_dtypes,
          err_msg=err_msg,
      )

    if dataclasses.is_dataclass(x):
      self.assertIs(type(y), type(x))
      for field in dataclasses.fields(x):  # pytype: disable=module-attr
        key = field.name
        x_value, y_value = getattr(x, key), getattr(y, key)
        is_pytree_node = field.metadata.get('pytree_node', True)
        if is_pytree_node:
          assert_close(x_value, y_value)
        else:
          self.assertEqual(x_value, y_value, key)
    else:
      assert_close(x, y)


# JAX MD specific utilities.


def update_test_tolerance(f32_tol=5e-3, f64_tol=1e-5):
  _DEFAULT_TOLERANCE[onp.dtype(onp.float32)] = f32_tol
  _DEFAULT_TOLERANCE[onp.dtype(onp.float64)] = f64_tol


def _load_silica_data(filename: str) -> jnp.ndarray:
  filename = FLAGS.test_srcdir + filename
  with open(filename, 'rb') as f:
    return jnp.array(onp.load(f))


def load_silica_data() -> jnp.ndarray:
  try:
    filename = 'tests/data/silica_positions.npy'
    return _load_silica_data(filename)
  except FileNotFoundError:
    filename = '/google3/third_party/py/jax_md/tests/data/silica_positions.npy'
    return _load_silica_data(filename)


def _load_elasticity_test_data(filename, spatial_dimension, dtype, index):
  filename = FLAGS.test_srcdir + filename
  ds = nc.Dataset(filename)
  cijkl = jnp.array(ds.variables['Cijkl'][index], dtype=dtype)
  R = jnp.array(ds.variables['pos'][index], dtype=dtype)
  R = jnp.reshape(R, (R.shape[0]//spatial_dimension,spatial_dimension))
  sigma = 2. * jnp.array(ds.variables['rad'][index], dtype=dtype)
  box = jnp.array(ds.variables['box'][index], dtype=dtype)
  box = jnp.reshape(box, (spatial_dimension,spatial_dimension))
  return cijkl, R, sigma, box


def load_elasticity_test_data(spatial_dimension, low_pressure, dtype, index):
  try:
    if low_pressure:
      if spatial_dimension == 2:
        fn = 'tests/data/2d_polyuniform_N64_Lp-4.0.nc'
      else:
        fn = 'tests/data/3d_bi_N128_Lp-4.0.nc'
    else:
      if spatial_dimension == 2:
        fn = 'tests/data/2d_polyuniform_N64_Lp-1.0.nc'
      else:
        fn = 'tests/data/3d_bi_N128_Lp-1.0.nc'
    return _load_elasticity_test_data(fn, spatial_dimension, dtype, index)
  except FileNotFoundError:
    if low_pressure:
      if spatial_dimension == 2:
        fn = '/google3/third_party/py/jax_md/tests/data/2d_polyuniform_N64_Lp-4.0.nc'
      else:
        fn = '/google3/third_party/py/jax_md/tests/data/3d_bi_N128_Lp-4.0.nc'
    else:
      if spatial_dimension == 2:
        fn = '/google3/third_party/py/jax_md/tests/data/2d_polyuniform_N64_Lp-1.0.nc'
      else:
        fn = '/google3/third_party/py/jax_md/tests/data/3d_bi_N128_Lp-1.0.nc'
    return _load_elasticity_test_data(fn, spatial_dimension, dtype, index)


@dataclasses.dataclass
class JammedTestState:
  fractional_position: jnp.ndarray
  real_position: jnp.ndarray
  species: jnp.ndarray
  sigma: jnp.ndarray
  box: jnp.ndarray
  energy: jnp.ndarray
  pressure: jnp.ndarray


def _load_jammed_state(filename: str, dtype) -> JammedTestState:
  filename = FLAGS.test_srcdir + filename
  with open(filename, 'rb') as f:
    return JammedTestState(
        fractional_position=onp.load(f).astype(dtype),
        real_position=onp.load(f).astype(dtype),
        species=onp.load(f),
        sigma=onp.load(f).astype(dtype),
        box=onp.load(f).astype(dtype),
        energy=onp.load(f).astype(dtype),
        pressure=onp.load(f).astype(dtype),  # pytype: disable=wrong-keyword-args
    )


def load_jammed_state(filename: str, dtype) -> JammedTestState:
  try:
    full_filename = f'tests/data/{filename}'
    return _load_jammed_state(full_filename, dtype)
  except FileNotFoundError:
    full_filename = f'/google3/third_party/py/jax_md/tests/data/{filename}'
    return _load_jammed_state(full_filename, dtype)


def load_lammps_stress_data(dtype):
  def parse_state(filename):
    filename = FLAGS.test_srcdir + filename
    with open(filename) as f:
      data = f.read()
    data = data.split('\n')
    t = int(data[1])
    n = int(data[3])
    box = float(data[5].split(' ')[-1])
    R = []
    V = []
    for l in data[9:-1]:
      R += [[float(xx) for xx in l.split(' ')[:3]]]
      V += [[float(xx) for xx in l.split(' ')[3:]]]
    return f32(box), jnp.array(R, dtype), jnp.array(V, dtype)

  def parse_results(filename):
    filename = FLAGS.test_srcdir + filename
    with open(filename) as f:
      data = f.read()
    data = [[float(dd) for dd in d.split(' ') if dd != ' ' and dd != '']
            for d in data.split('\n')]
    step = jnp.array([int(d[0]) for d in data if len(d) > 0])
    Es = jnp.array([d[1] for d in data if len(d) > 0], dtype)
    C = jnp.array([d[2:] for d in data if len(d) > 0], dtype)
    C = jnp.array([[C[0, 0], C[0, 3], C[0, 4]],
                   [C[0, 3], C[0, 1], C[0, 5]],
                   [C[0, 4], C[0, 5], C[0, 2]]], dtype)
    return Es[0], C

  try:
    return (parse_state('tests/data/lammps_lj_stress_test_states'),
            parse_results('tests/data/lammps_lj_stress_test'))
  except FileNotFoundError:
    return (parse_state('/google3/third_party/py/jax_md/tests/data/'
                        'lammps_lj_stress_test_states'),
            parse_results('/google3/third_party/py/jax_md/tests/data/'
                          'lammps_lj_stress_test'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for constructing various interpolating functions.

This code was adapted from the way learning rate schedules are are built in JAX.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import jax.numpy as np
from scipy.interpolate import splrep, PPoly

from jax_md import util


# Typing

f32 = util.f32
f64 = util.f64

#


def constant(f):
  def schedule(unused_t):
    return f
  return schedule


def canonicalize(scalar_or_schedule_fun):
  if callable(scalar_or_schedule_fun):
    return scalar_or_schedule_fun
  elif np.ndim(scalar_or_schedule_fun) == 0:
    return constant(scalar_or_schedule_fun)
  else:
    raise TypeError(type(scalar_or_schedule_fun))


def spline(y, dx, degree=3):
  """Spline fit a given scalar function.

  Args:
    y: The values of the scalar function evaluated on points starting at zero
    with the interval dx.
    dx: The interval at which the scalar function is evaluated.
    degree: Polynomial degree of the spline fit.

  Returns:
    A function that computes the spline function.
  """
  num_points = len(y)
  dx = f32(dx)
  x = np.arange(num_points, dtype=f32) * dx
  # Create a spline fit using the scipy function.
  fn = splrep(x, y, s=0, k=degree)  # Turn off smoothing by setting s to zero.
  params = PPoly.from_spline(fn)
  # Store the coefficients of the spline fit to an array.
  coeffs = np.array(params.c)
  def spline_fn(x):
    """Evaluates the spline fit for values of x."""
    ind = np.array(x / dx, dtype=np.int64)
    # The spline is defined for x values between 0 and largest value of y. If x
    # is outside this domain, truncate its ind value to within the domain.
    truncated_ind = np.array(
        np.where(ind < num_points, ind, num_points - 1), np.int64)
    truncated_ind = np.array(
        np.where(truncated_ind >= 0, truncated_ind, 0), np.int64)
    result = np.array(0, x.dtype)
    dX = x - np.array(ind, np.float32) * dx
    for i in range(degree + 1):  # sum over the polynomial terms up to degree.
      result = result + np.array(coeffs[degree - i, truncated_ind + 2],
                                 x.dtype) * dX ** np.array(i, x.dtype)
    # For x values that are outside the domain of the spline fit, return zeros.
    result = np.where(ind < num_points, result, np.array(0.0, x.dtype))
    return result
  return spline_fn
